import pytest
import regex
from rdflib import RDF
from rdflib import XSD
from rdflib import BNode
from rdflib import Literal
from rdflib import Namespace
from rdflib import URIRef

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import KNORA_API
from dsp_tools.commands.xmlupload.make_rdf_graph.make_values import _make_one_value_graph
from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryBoolean
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryColor
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryDate
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryDecimal
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryGeometry
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryGeoname
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryInt
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryInterval
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryLink
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryList
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryRichtext
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediarySimpleText
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryTime
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryUri
from dsp_tools.commands.xmlupload.models.intermediary.values import IntervalFloats
from dsp_tools.commands.xmlupload.models.lookup_models import IRILookups
from dsp_tools.commands.xmlupload.models.lookup_models import JSONLDContext
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.permission import PermissionValue
from dsp_tools.models.exceptions import BaseError
from dsp_tools.utils.date_util import Calendar
from dsp_tools.utils.date_util import Date
from dsp_tools.utils.date_util import Era
from dsp_tools.utils.date_util import SingleDate

ONTO = Namespace("http://0.0.0.0:3333/ontology/9999/onto/v2#")

PERMISSION_LITERAL = Literal("CR knora-admin:ProjectAdmin", datatype=XSD.string)
RES_ONE_URI = URIRef("http://rdfh.ch/9999/res_one")


OPEN_PERMISSION = Permissions({PermissionValue.CR: ["knora-admin:ProjectAdmin"]})


def absolute_iri(prop: str) -> str:
    return f"http://0.0.0.0:3333/ontology/9999/onto/v2#{prop}"


@pytest.fixture
def lookups() -> IRILookups:
    return IRILookups(
        project_iri=URIRef("http://rdfh.ch/9999/project"),
        id_to_iri=IriResolver({"res_one": "http://rdfh.ch/9999/res_one"}),
        jsonld_context=JSONLDContext({}),
    )


class TestMakeOneValueGraphSuccess:
    def test_boolean(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediaryBoolean(True, absolute_iri("isTrueOrFalse"), None, OPEN_PERMISSION)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 4
        val_bn = next(result.objects(res_bn, ONTO.isTrueOrFalse))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.BooleanValue
        value = next(result.objects(val_bn, KNORA_API.booleanValueAsBoolean))
        assert value == Literal(True, datatype=XSD.boolean)
        permissions = next(result.objects(val_bn, KNORA_API.hasPermissions))
        assert permissions == PERMISSION_LITERAL

    def test_color(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediaryColor("#5d1f1e", absolute_iri("hasColor"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasColor))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.ColorValue
        value = next(result.objects(val_bn, KNORA_API.colorValueAsColor))
        assert value == Literal("#5d1f1e", datatype=XSD.string)

    def test_decimal(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediaryDecimal(2.718281828459, absolute_iri("hasDecimal"), "Eulersche Zahl", None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 4
        val_bn = next(result.objects(res_bn, ONTO.hasDecimal))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.DecimalValue
        value = next(result.objects(val_bn, KNORA_API.decimalValueAsDecimal))
        assert value == Literal("2.718281828459", datatype=XSD.decimal)
        comment = next(result.objects(val_bn, KNORA_API.valueHasComment))
        assert comment == Literal("Eulersche Zahl", datatype=XSD.string)

    def test_geometry(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediaryGeometry(
            '{"status": "active", "type": "polygon", "lineWidth": 5, '
            '"points": [{"x": 0.4, "y": 0.6}, {"x": 0.5, "y": 0.9}, {"x": 0.8, "y": 0.9}, {"x": 0.7, "y": 0.6}]}',
            absolute_iri("hasGeometry"),
            None,
            None,
        )
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasGeometry))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.GeomValue
        value = next(result.objects(val_bn, KNORA_API.geometryValueAsGeometry))
        assert isinstance(value, Literal)

    def test_geoname(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediaryGeoname("5416656", absolute_iri("hasGeoname"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasGeoname))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.GeonameValue
        value = next(result.objects(val_bn, KNORA_API.geonameValueAsGeonameCode))
        assert value == Literal("5416656", datatype=XSD.string)

    def test_integer(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediaryInt(1, absolute_iri("hasInteger"), "comment", None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 4
        val = next(result.subjects(KNORA_API.intValueAsInt, Literal("1", datatype=XSD.integer)))
        assert next(result.objects(val, RDF.type)) == KNORA_API.IntValue
        assert next(result.subjects(ONTO.hasInteger, val)) == res_bn
        comment = next(result.objects(val, KNORA_API.valueHasComment))
        assert comment == Literal("comment", datatype=XSD.string)

    def test_time(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediaryTime("2019-10-23T13:45:12.01-14:00", absolute_iri("hasTime"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasTime))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.TimeValue
        value = next(result.objects(val_bn, KNORA_API.timeValueAsTimeStamp))
        assert value == Literal("2019-10-23T13:45:12.01-14:00", datatype=XSD.dateTimeStamp)

    def test_uri(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediaryUri("https://dasch.swiss", absolute_iri("hasUri"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasUri))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.UriValue
        value = next(result.objects(val_bn, KNORA_API.uriValueAsUri))
        assert value == Literal("https://dasch.swiss", datatype=XSD.anyURI)

    def test_list(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediaryList("http://rdfh.ch/9999/node", absolute_iri("hasListItem"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasListItem))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.ListValue
        value = next(result.objects(val_bn, KNORA_API.listValueAsListNode))
        assert value == URIRef("http://rdfh.ch/9999/node")

    def test_resptr(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediaryLink("res_one", absolute_iri("hasResource"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasResourceValue))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.LinkValue
        value = next(result.objects(val_bn, KNORA_API.linkValueHasTargetIri))
        assert value == RES_ONE_URI

    def test_simpletext(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediarySimpleText("Text", absolute_iri("hasSimpleText"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasSimpleText))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.TextValue
        value = next(result.objects(val_bn, KNORA_API.valueAsString))
        assert value == Literal("Text", datatype=XSD.string)

    def test_richtext(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediaryRichtext(
            FormattedTextValue("Text"), absolute_iri("hasRichtext"), None, OPEN_PERMISSION, resource_references=set()
        )
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 5
        val_bn = next(result.objects(res_bn, ONTO.hasRichtext))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.TextValue
        value = next(result.objects(val_bn, KNORA_API.textValueAsXml))
        # FormattedTextValue adds a newline after the xml declaration
        assert value == Literal('<?xml version="1.0" encoding="UTF-8"?>\n<text>Text</text>', datatype=XSD.string)
        mapping = next(result.objects(val_bn, KNORA_API.textValueHasMapping))
        assert mapping == URIRef("http://rdfh.ch/standoff/mappings/StandardMapping")
        permissions = next(result.objects(val_bn, KNORA_API.hasPermissions))
        assert permissions == PERMISSION_LITERAL

    def test_richtext_with_reference(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        text = 'Comment with <a class="salsah-link" href="IRI:res_one:IRI">link to res_one'
        prop = IntermediaryRichtext(
            FormattedTextValue(text), absolute_iri("hasRichtext"), None, None, resource_references=set("res_one")
        )
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 4
        val_bn = next(result.objects(res_bn, ONTO.hasRichtext))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.TextValue
        value = next(result.objects(val_bn, KNORA_API.textValueAsXml))
        expected_text = (
            '<?xml version="1.0" encoding="UTF-8"?>\n'  # FormattedTextValue adds a newline after the xml declaration
            '<text>Comment with <a class="salsah-link" href="http://rdfh.ch/9999/res_one">'
            "link to res_one</text>"
        )
        assert value == Literal(expected_text, datatype=XSD.string)
        mapping = next(result.objects(val_bn, KNORA_API.textValueHasMapping))
        assert mapping == URIRef("http://rdfh.ch/standoff/mappings/StandardMapping")

    def test_date(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        date = Date(
            calendar=Calendar.GREGORIAN,
            start=SingleDate(era=Era.AD, year=476, month=9, day=4),
            end=SingleDate(era=Era.AD, year=477, month=None, day=None),
        )
        prop = IntermediaryDate(date, absolute_iri("hasDate"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 9
        val_bn = next(result.objects(res_bn, ONTO.hasDate))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.DateValue
        calendar = next(result.objects(val_bn, KNORA_API.dateValueHasCalendar))
        assert calendar == Literal("GREGORIAN", datatype=XSD.string)
        start_day = next(result.objects(val_bn, KNORA_API.dateValueHasStartDay))
        assert start_day == Literal("4", datatype=XSD.integer)
        start_month = next(result.objects(val_bn, KNORA_API.dateValueHasStartMonth))
        assert start_month == Literal("9", datatype=XSD.integer)
        start_year = next(result.objects(val_bn, KNORA_API.dateValueHasStartYear))
        assert start_year == Literal("476", datatype=XSD.integer)
        start_era = next(result.objects(val_bn, KNORA_API.dateValueHasStartEra))
        assert start_era == Literal("AD", datatype=XSD.string)
        end_year = next(result.objects(val_bn, KNORA_API.dateValueHasEndYear))
        assert end_year == Literal("477", datatype=XSD.integer)
        start_era = next(result.objects(val_bn, KNORA_API.dateValueHasEndEra))
        assert start_era == Literal("AD", datatype=XSD.string)

    def test_interval(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        interval = IntervalFloats(0.1, 0.234)
        prop = IntermediaryInterval(interval, "http://api.knora.org/ontology/knora-api/v2#hasSegmentBounds", None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 4
        val_bn = next(result.objects(res_bn, KNORA_API.hasSegmentBounds))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.IntervalValue
        start = next(result.objects(val_bn, KNORA_API.intervalValueHasStart))
        assert start == Literal("0.1", datatype=XSD.decimal)
        end = next(result.objects(val_bn, KNORA_API.intervalValueHasEnd))
        assert end == Literal("0.234", datatype=XSD.decimal)

    def test_segment_of_video(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediaryLink("res_one", "http://api.knora.org/ontology/knora-api/v2#isVideoSegmentOf", None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, KNORA_API.isVideoSegmentOfValue))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.LinkValue
        value = next(result.objects(val_bn, KNORA_API.linkValueHasTargetIri))
        assert value == RES_ONE_URI

    def test_segment_of_audio(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = IntermediaryLink("res_one", "http://api.knora.org/ontology/knora-api/v2#isAudioSegmentOf", None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, KNORA_API.isAudioSegmentOfValue))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.LinkValue
        value = next(result.objects(val_bn, KNORA_API.linkValueHasTargetIri))
        assert value == RES_ONE_URI


def test_link_target_not_found(lookups: IRILookups) -> None:
    res_bn = BNode()
    prop = IntermediaryLink("non_existing", absolute_iri("hasResource"), None, None)
    err_str = regex.escape(
        (
            "Could not find the ID non_existing in the id2iri mapping. "
            "This is probably because the resource 'non_existing' could not be created. "
            "See warnings.log for more information."
        )
    )
    with pytest.raises(BaseError, match=err_str):
        _make_one_value_graph(prop, res_bn, lookups)


def test_richtext_with_reference_not_found(lookups: IRILookups) -> None:
    res_bn = BNode()
    text = 'Comment with <a class="salsah-link" href="IRI:nonExisingReference:IRI">link to res_one'
    prop = IntermediaryRichtext(
        FormattedTextValue(text),
        absolute_iri("hasRichtext"),
        None,
        None,
        resource_references=set("nonExisingReference"),
    )
    err_str = regex.escape("Internal ID 'nonExisingReference' could not be resolved to an IRI")
    with pytest.raises(BaseError, match=err_str):
        _make_one_value_graph(prop, res_bn, lookups)



import pytest

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue


def test_find_ids_referenced_in_salsah_links_one_link() -> None:
    one_link_KnoraStandoffXml = FormattedTextValue(
        xmlstr=(
            '<resource label="r1_label" restype="r1_restype" id="r1_id" permissions="open">'
            '<text-prop name=":hasRichtext"><text permissions="open" encoding="xml">'
            '<a class="salsah-link" href="IRI:r2_id:IRI">r2_id</a>'
            "</text></text-prop></resource>"
        )
    )
    returned_set = one_link_KnoraStandoffXml.find_internal_ids()
    assert returned_set == {"r2_id"}


def test_find_ids_referenced_in_salsah_links_three_links() -> None:
    three_link_KnoraStandoffXml = FormattedTextValue(
        xmlstr=(
            '<resource label="r1_label" restype="r1_restype" id="r1_id" permissions="open">'
            '<text-prop name=":hasRichtext"><text permissions="open" encoding="xml">'
            '<a class="salsah-link" href="IRI:r2_id:IRI">r2_id</a>This is normal text'
            '<a class="salsah-link" href="IRI:r3_id:IRI">r3_id</a>'
            '<a class="salsah-link" href="IRI:r2_id:IRI">r2_id</a>'
            "</text></text-prop></resource>"
        )
    )
    returned_set = three_link_KnoraStandoffXml.find_internal_ids()
    assert returned_set == {"r2_id", "r3_id"}


def test__replace_internal_ids_with_iris_one_link() -> None:
    resolver = IriResolver({"r1_id": "r1_iri", "r2_id": "r2_iri", "r3_id": "r3_iri"})
    one_link_KnoraStandoffXml = FormattedTextValue(
        xmlstr=(
            '<resource label="r1_label" restype="r1_restype" id="r1_id" permissions="open">'
            '<text-prop name=":hasRichtext"><text permissions="open" encoding="xml">'
            '<a class="salsah-link" href="IRI:r2_id:IRI">r2_id</a>'
            "</text></text-prop></resource>"
        )
    )
    returned_instance = one_link_KnoraStandoffXml.with_iris(resolver)
    expected_str = (
        '<resource label="r1_label" restype="r1_restype" id="r1_id" permissions="open">'
        '<text-prop name=":hasRichtext"><text permissions="open" encoding="xml">'
        '<a class="salsah-link" href="r2_iri">r2_id</a>'
        "</text></text-prop></resource>"
    )
    assert returned_instance.xmlstr == expected_str


def test__replace_internal_ids_with_iris_three_links() -> None:
    resolver = IriResolver({"r1_id": "r1_iri", "r2_id": "r2_iri", "r3_id": "r3_iri"})
    three_link_KnoraStandoffXml = FormattedTextValue(
        xmlstr=(
            '<resource label="r1_label" restype="r1_restype" id="r1_id" permissions="open">'
            '<text-prop name=":hasRichtext"><text permissions="open" encoding="xml">'
            '<a class="salsah-link" href="IRI:r2_id:IRI">r2_id</a>This is normal text'
            '<a class="salsah-link" href="IRI:r3_id:IRI">r3_id</a>'
            '<a class="salsah-link" href="IRI:r2_id:IRI">r2_id</a>'
            "</text></text-prop></resource>"
        )
    )
    returned_instance = three_link_KnoraStandoffXml.with_iris(resolver)
    expected_str = (
        '<resource label="r1_label" restype="r1_restype" id="r1_id" permissions="open">'
        '<text-prop name=":hasRichtext"><text permissions="open" encoding="xml">'
        '<a class="salsah-link" href="r2_iri">r2_id</a>This is normal text'
        '<a class="salsah-link" href="r3_iri">r3_id</a>'
        '<a class="salsah-link" href="r2_iri">r2_id</a>'
        "</text></text-prop></resource>"
    )
    assert returned_instance.xmlstr == expected_str


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

import warnings
from abc import ABC
from abc import abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

import pandas as pd
from rdflib import Graph

from dsp_tools.commands.validate_data.models.validation import UnexpectedComponent
from dsp_tools.models.custom_warnings import DspToolsUserWarning

LIST_SEPARATOR = "\n    - "
GRAND_SEPARATOR = "\n\n----------------------------\n"


@dataclass
class UnknownClassesInData:
    unknown_classes: set[str]
    classes_onto: set[str]

    def get_msg(self) -> str:
        if unknown := self._get_unknown_ontos_msg():
            return unknown
        return self._get_unknown_classes_msg()

    def _get_unknown_ontos_msg(self) -> str:
        def split_prefix(relative_iri: str) -> str:
            return relative_iri.split(":")[0]

        used_ontos = set(split_prefix(x) for x in self.unknown_classes)
        exising_ontos = set(split_prefix(x) for x in self.classes_onto)
        msg = ""
        if unknown := used_ontos - exising_ontos:
            msg = (
                f"Your data uses ontologies that don't exist in the database.\n"
                f"The following ontologies that are used in the data are unknown: {', '.join(exising_ontos)}"
                f"The following ontologies are uploaded: {', '.join(unknown)}\n"
            )
        return msg

    def _get_unknown_classes_msg(self) -> str:
        unknown_classes = sorted(list(self.unknown_classes))
        known_classes = sorted(list(self.classes_onto))
        return (
            f"Your data uses resource classes that do not exist in the ontologies in the database.\n"
            f"The following classes that are used in the data are unknown: {', '.join(unknown_classes)}\n"
            f"The following classes exist in the uploaded ontologies: {', '.join(known_classes)}\n"
        )


@dataclass
class UnexpectedResults:
    components: list[UnexpectedComponent]

    def save_inform_user(self, results_graph: Graph, shacl: Graph, data: Graph) -> None:
        cwdr = Path.cwd()
        prefix = f"{datetime.now()!s}_"
        unique_components = list(set(x.component_type for x in self.components))
        components = sorted(unique_components)
        msg = (
            f"Unexpected violations were found in the validation results:"
            f"{LIST_SEPARATOR}{LIST_SEPARATOR.join(components)}\n"
            f"Please contact the development team with the files starting with the timestamp '{prefix}' "
            f"in the directory '{cwdr}'."
        )
        save_path = cwdr / f"{prefix}validation_result.ttl"
        results_graph.serialize(save_path)
        shacl_p = cwdr / f"{prefix}shacl.ttl"
        shacl.serialize(shacl_p)
        data_p = cwdr / f"{prefix}data.ttl"
        data.serialize(data_p)
        warnings.warn(DspToolsUserWarning(msg))


@dataclass
class AllProblems:
    problems: list[InputProblem]
    unexpected_results: UnexpectedResults | None

    def get_msg(self, file_path: Path) -> str:
        coll = self._make_collection()
        title_msg = f"\nDuring the validation of the data {len(self.problems)} errors were found:\n\n"
        if len(self.problems) > 50:
            out_path = file_path.parent / f"{file_path.stem}_validation_errors.csv"
            self._save_as_csv(out_path)
            out_message = (
                title_msg + f"Due to the large number or errors, the validation errors were saved at:\n{out_path}"
            )
        else:
            msg = [x.get_msg() for x in coll]
            out_message = title_msg + GRAND_SEPARATOR.join(msg)
        return out_message

    def _make_collection(self) -> list[ResourceProblemCollection]:
        d = defaultdict(list)
        for e in self.problems:
            d[e.res_id].append(e)
        collection_list = []
        for k, v in d.items():
            collection_list.append(ResourceProblemCollection(k, v))
        return sorted(collection_list, key=lambda x: x.res_id)

    def _save_as_csv(self, out_path: Path) -> None:
        all_problems = [x.to_dict() for x in self.problems]
        df = pd.DataFrame.from_records(all_problems)
        df = df.sort_values(by=["Resource Type", "Resource ID", "Property"])
        df.to_csv(out_path, index=False)


@dataclass
class ResourceProblemCollection:
    res_id: str
    problems: list[InputProblem]

    def get_msg(self) -> str:
        prop_msg = self._msg_for_properties()
        return f"Resource ID: {self.res_id} | Resource Type: {self.problems[0].res_type}\n{prop_msg}"

    def _msg_for_properties(self) -> str:
        grouped = self._make_collection()
        out_list = []
        for prop_name, problem in grouped.items():
            problem_list = [x.get_msg() for x in problem]
            out_list.append((prop_name, f"{prop_name}{LIST_SEPARATOR}{LIST_SEPARATOR.join(problem_list)}"))
        sorted_list = sorted(out_list, key=lambda x: x[0])
        return "\n".join([x[1] for x in sorted_list])

    def _make_collection(self) -> dict[str, list[InputProblem]]:
        grouped_dict = defaultdict(list)
        for problem in self.problems:
            grouped_dict[problem.sort_value()].append(problem)
        return grouped_dict

    def sort_value(self) -> str:
        return self.res_id


@dataclass
class InputProblem(ABC):
    res_id: str
    res_type: str
    prop_name: str

    @property
    @abstractmethod
    def problem(self) -> str: ...

    @abstractmethod
    def get_msg(self) -> str: ...

    @abstractmethod
    def to_dict(self) -> dict[str, str]: ...

    def _base_dict(self) -> dict[str, str]:
        return {
            "Resource Type": self.res_type,
            "Resource ID": self.res_id,
            "Property": self.prop_name,
            "Problem": self.problem,
        }

    @abstractmethod
    def sort_value(self) -> str: ...


#######################
# Generic Problem


@dataclass
class GenericProblem(InputProblem):
    results_message: str
    actual_content: str

    @property
    def problem(self) -> str:
        return self.results_message

    def get_msg(self) -> str:
        return f"{self.problem} | Content: '{self.actual_content}'"

    def to_dict(self) -> dict[str, str]:
        problm_dict = self._base_dict()
        problm_dict["Actual"] = self.actual_content
        return problm_dict

    def sort_value(self) -> str:
        return self.prop_name


#######################
# Cardinality Violation


@dataclass
class MaxCardinalityProblem(InputProblem):
    expected_cardinality: str

    @property
    def problem(self) -> str:
        return "Maximum Cardinality Violation"

    def get_msg(self) -> str:
        return f"{self.problem} | Expected Cardinality: {self.expected_cardinality}"

    def to_dict(self) -> dict[str, str]:
        problm_dict = self._base_dict()
        problm_dict["Expected"] = f"Cardinality: {self.expected_cardinality}"
        return problm_dict

    def sort_value(self) -> str:
        return self.prop_name


@dataclass
class MinCardinalityProblem(InputProblem):
    expected_cardinality: str

    @property
    def problem(self) -> str:
        return "Minimum Cardinality Violation"

    def get_msg(self) -> str:
        return f"{self.problem} | Expected Cardinality: {self.expected_cardinality}"

    def to_dict(self) -> dict[str, str]:
        problm_dict = self._base_dict()
        problm_dict["Expected"] = f"Cardinality: {self.expected_cardinality}"
        return problm_dict

    def sort_value(self) -> str:
        return self.prop_name


@dataclass
class NonExistentCardinalityProblem(InputProblem):
    @property
    def problem(self) -> str:
        return "The resource class does not have a cardinality for this property."

    def get_msg(self) -> str:
        return self.problem

    def to_dict(self) -> dict[str, str]:
        return self._base_dict()

    def sort_value(self) -> str:
        return self.prop_name


@dataclass
class FileValueNotAllowedProblem(InputProblem):
    @property
    def problem(self) -> str:
        return "A file was added to the resource. This resource type must not have a file."

    def get_msg(self) -> str:
        return self.problem

    def to_dict(self) -> dict[str, str]:
        return self._base_dict()

    def sort_value(self) -> str:
        return self.prop_name


#######################
# Value Type Violation


@dataclass
class ValueTypeProblem(InputProblem):
    actual_type: str
    expected_type: str

    @property
    def problem(self) -> str:
        return "Value Type Mismatch"

    def get_msg(self) -> str:
        return f"{self.problem}, Actual Type: {self.actual_type} | Expected Type: {self.expected_type}"

    def to_dict(self) -> dict[str, str]:
        problm_dict = self._base_dict()
        problm_dict["Expected"] = self.expected_type
        problm_dict["Actual"] = self.actual_type
        return problm_dict

    def sort_value(self) -> str:
        return self.prop_name


#######################
# Content Violation


@dataclass
class ContentRegexProblem(InputProblem):
    expected_format: str
    actual_content: str | None

    @property
    def problem(self) -> str:
        return "Wrong Content Format"

    def get_msg(self) -> str:
        msg = f"{self.problem}, Expected Format: {self.expected_format}"
        if self.actual_content:
            msg += f" | Content: '{self._short_content()}'"
        return msg

    def to_dict(self) -> dict[str, str]:
        problm_dict = self._base_dict()
        problm_dict["Expected"] = self.expected_format
        if self.actual_content:
            problm_dict["Content"] = self._short_content()
        return problm_dict

    def _short_content(self) -> str:
        if not self.actual_content:
            return ""
        if len(self.actual_content) > 15:
            return f"{self.actual_content[:15]}[...]"
        return self.actual_content

    def sort_value(self) -> str:
        return self.prop_name


@dataclass
class LinkTargetTypeMismatchProblem(InputProblem):
    link_target_id: str
    actual_type: str
    expected_type: str

    @property
    def problem(self) -> str:
        return "Linked Resource Type Mismatch"

    def get_msg(self) -> str:
        return (
            f"{self.problem}, Target Resource ID: '{self.link_target_id}' "
            f"Actual Type: '{self.actual_type}' | Expected Resource Type: '{self.expected_type}'"
        )

    def to_dict(self) -> dict[str, str]:
        problm_dict = self._base_dict()
        problm_dict["Expected"] = self.expected_type
        problm_dict["Actual"] = self.actual_type
        problm_dict["Content"] = self.link_target_id
        return problm_dict

    def sort_value(self) -> str:
        return self.prop_name


@dataclass
class LinkedResourceDoesNotExistProblem(InputProblem):
    link_target_id: str

    @property
    def problem(self) -> str:
        return "Linked Resource does not Exist"

    def get_msg(self) -> str:
        return f"{self.problem}, Target Resource ID: '{self.link_target_id}'"

    def to_dict(self) -> dict[str, str]:
        problm_dict = self._base_dict()
        problm_dict["Actual"] = self.link_target_id
        return problm_dict

    def sort_value(self) -> str:
        return self.prop_name


@dataclass
class DuplicateValueProblem(InputProblem):
    actual_content: str

    @property
    def problem(self) -> str:
        return "Value is duplicated"

    def get_msg(self) -> str:
        return f"{self.problem}, Content: '{self.actual_content}'"

    def to_dict(self) -> dict[str, str]:
        problm_dict = self._base_dict()
        problm_dict["Content"] = self.actual_content
        return problm_dict

    def sort_value(self) -> str:
        return self.prop_name


@dataclass
class FileValueProblem(InputProblem):
    expected: str

    @property
    def problem(self) -> str:
        return self.expected

    def get_msg(self) -> str:
        return f"{self.problem}"

    def to_dict(self) -> dict[str, str]:
        problm_dict = self._base_dict()
        problm_dict["Expected"] = self.expected
        return problm_dict

    def sort_value(self) -> str:
        return self.prop_name



import contextlib
import glob
import json
import shutil
import unittest
from pathlib import Path
from typing import Any
from typing import Optional
from typing import Union
from typing import cast

import regex

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.id2iri import id2iri
from dsp_tools.commands.project.create.project_create import create_project
from dsp_tools.commands.project.get import get_project
from dsp_tools.commands.xmlupload.xmlupload import xmlupload

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)


class TestCreateGetXMLUpload(unittest.TestCase):
    """Test if the commands "create", "get", and "xmlupload" work together as expected."""

    creds = ServerCredentials(
        user="root@example.com", password="test", server="http://0.0.0.0:3333", dsp_ingest_url="http://0.0.0.0:3340"
    )
    imgdir = "."
    test_project_systematic_file = Path("testdata/json-project/test-project-systematic.json")
    test_data_systematic_file = Path("testdata/xml-data/test-data-systematic.xml")
    cwd = Path("cwd")
    testdata_tmp = Path("testdata/tmp")

    @classmethod
    def setUpClass(cls) -> None:
        """Is executed once before the methods of this class are run"""
        cls.testdata_tmp.mkdir(exist_ok=True)
        cls.cwd.mkdir(exist_ok=True)

    @classmethod
    def tearDownClass(cls) -> None:
        """Is executed after the methods of this class have all run through"""
        shutil.rmtree(cls.testdata_tmp)
        shutil.rmtree(cls.cwd)
        for f in Path().glob("*id2iri_*.json"):
            f.unlink()

    def test_create_project(self) -> None:
        """Test if the systematic JSON project file can be uploaded without producing an error on its way"""
        success = create_project(
            project_file_as_path_or_parsed=self.test_project_systematic_file.absolute(),
            creds=self.creds,
            verbose=True,
        )
        self.assertTrue(success)

    def test_xml_upload_incremental(self) -> None:
        """
        Test if the systematic XML data file can be uploaded without producing an error on its way,
        and if the 'id2iri' replacement works, so that the 2nd upload works.
        """
        success = xmlupload(
            input_file=self.test_data_systematic_file,
            creds=self.creds,
            imgdir=self.imgdir,
        )
        self.assertTrue(success)

        mapping_file = self._get_most_recent_glob_match("*_id2iri_mapping_*.json")
        second_xml_file_orig = Path("testdata/id2iri/test-id2iri-data.xml")
        success = id2iri(
            xml_file=str(second_xml_file_orig),
            json_file=str(mapping_file),
        )
        mapping_file.unlink()
        self.assertTrue(success)

        second_xml_file_replaced = self._get_most_recent_glob_match(f"{second_xml_file_orig.stem}_replaced_*.xml")
        success = xmlupload(
            input_file=second_xml_file_replaced,
            creds=self.creds,
            imgdir=self.imgdir,
        )
        second_xml_file_replaced.unlink()
        self.assertListEqual(list(Path(self.cwd).glob("stashed_*_properties_*.txt")), [])
        self.assertTrue(success)

    def test_get_project(self) -> None:
        """
        Retrieve the systematic JSON project file with the "get" command,
        and check if the result is identical to the original file.
        """
        out_file = self.testdata_tmp / "_test-project-systematic.json"
        success = get_project(
            project_identifier="systematic-tp",
            outfile_path=str(out_file),
            creds=self.creds,
            verbose=True,
        )
        self.assertTrue(success)

        project_original = self._get_original_project()
        with open(out_file, encoding="utf-8") as f:
            project_returned = json.load(f)

        self._compare_project(project_original, project_returned)
        self._compare_groups(
            groups_original=project_original["project"].get("groups"),
            groups_returned=project_returned["project"].get("groups"),
        )
        self._compare_users(
            users_original=project_original["project"].get("users"),
            users_returned=project_returned["project"].get("users"),
            project_shortname=project_original["project"]["shortname"],
        )
        self._compare_lists(
            lists_original=project_original["project"].get("lists"),
            lists_returned=project_returned["project"].get("lists"),
        )

        for file in [project_original, project_returned]:
            file["project"]["ontologies"] = sorted(file["project"]["ontologies"], key=lambda x: cast(str, x["name"]))
        for onto_original, onto_returned in zip(
            project_original["project"]["ontologies"],
            project_returned["project"]["ontologies"],
        ):
            self._compare_properties(
                properties_original=onto_original["properties"],
                properties_returned=onto_returned["properties"],
                onto_name=onto_original["name"],
            )
            self._compare_resources(
                resources_original=onto_original["resources"],
                resources_returned=onto_returned["resources"],
                onto_name=onto_original["name"],
            )

    def _get_original_project(self) -> dict[str, Any]:
        """
        Open the systematic JSON project file, expand all prefixes, and return the resulting Python object.
        The prefixes must be expanded because the "get" command returns full IRIs instead of prefixed IRIs,
        so the original prefixes won't be found in the returned file.

        Returns:
            the original project as a Python dictionary
        """
        with open(self.test_project_systematic_file, encoding="utf-8") as f:
            project_original_str = f.read()
            project_original = json.loads(project_original_str)

        for prefix, iri in project_original["prefixes"].items():
            project_original_str = regex.sub(rf'"{prefix}:(\w+?)"', rf'"{iri}\1"', project_original_str)
        project_original = json.loads(project_original_str)
        return cast(dict[str, Any], project_original)

    def _compare_project(
        self,
        project_original: dict[str, Any],
        project_returned: dict[str, Any],
    ) -> None:
        """
        Compare the basic metadata of 2 JSON project definitions.
        Fails with a message if the metadata are not identical.

        Args:
            project_original: original file
            project_returned: returned file
        """
        for field in ["shortcode", "shortname", "longname", "descriptions"]:
            orig = project_original["project"].get(field)
            ret = project_returned["project"].get(field)
            self.assertEqual(orig, ret, msg=f"Field '{field}' is not identical: original='{orig}', returned='{ret}'")

        orig_keywords = sorted(project_original["project"]["keywords"])
        ret_keywords = sorted(project_returned["project"]["keywords"])
        self.assertEqual(
            orig_keywords,
            ret_keywords,
            msg=f"Field keywords is not identical: original={orig_keywords}, returned={ret_keywords}",
        )

    def _compare_groups(
        self,
        groups_original: Optional[list[dict[str, Any]]],
        groups_returned: Optional[list[dict[str, Any]]],
    ) -> None:
        """
        Compare the "groups" section of the original JSON project definition
        with the "groups" section of the JSON returned by the "get" command.
        Fails with a message if the sections are not identical.

        Args:
            groups_original: "groups" section of the original file.
            groups_returned: "groups" section of the returned file.
        """
        # avoid mutable default argument
        groups_original = groups_original or []
        groups_returned = groups_returned or []

        # write default values into original, in case they were omitted (the "get" command writes all default values)
        for group in groups_original or []:
            if group.get("status") is None:
                group["status"] = True
            if group.get("selfjoin") is None:
                group["selfjoin"] = False

        # sort both lists
        groups_original = sorted(groups_original, key=lambda x: cast(str, x.get("name", "")))
        groups_returned = sorted(groups_returned, key=lambda x: cast(str, x.get("name", "")))
        if len(groups_original) != len(groups_returned):
            self.assertEqual(
                groups_original,
                groups_returned,
                msg="Returned number of groups is different from original number of groups.",
            )
        else:
            for orig, ret in zip(groups_original, groups_returned):
                self.assertEqual(
                    orig,
                    ret,
                    msg=f"Group with original name '{orig['name']}' and returned name '{ret['name']}' failed.",
                )

    def _compare_users(
        self,
        users_original: Optional[list[dict[str, Any]]],
        users_returned: Optional[list[dict[str, Any]]],
        project_shortname: str,
    ) -> None:
        """
        Compare the "users" section of the original JSON project definition
        with the "users" section of the JSON returned by the "get" command.
        Fails with a message if the sections are not identical.

        In order to do so, this method modifies the original as follows:
         - remove the users that are not in the current project
         - set all passwords to ""
         - add default values, in case they were omitted,
         - expand ":xyz" to "project_shortname:xyz"

        Args:
            users_original: "users" section of the original file.
            users_returned: "users" section of the returned file.
            project_shortname: shortname of the project
        """
        # avoid mutable default argument
        users_original = users_original or []
        users_returned = users_returned or []

        # remove users that are not in current project (they won't be returned by the "get" command)
        if users_original:
            current_project_membership_strings = [
                ":admin",
                ":member",
                f"{project_shortname}:admin",
                f"{project_shortname}:member",
            ]
            users_original = [
                u for u in users_original if any(p in u.get("projects", []) for p in current_project_membership_strings)
            ]

        # bring the "users" section of original into the form that is returned by the "get" command
        for user in users_original:
            index = users_original.index(user)
            # remove password
            users_original[index]["password"] = ""
            # write default values, in case they were omitted
            if user.get("groups") is None:
                users_original[index]["groups"] = []
            if user.get("status") is None:
                users_original[index]["status"] = True
            # expand ":xyz" to "project_shortname:xyz"
            if user.get("groups"):
                users_original[index]["groups"] = [regex.sub("^:", f"{project_shortname}:", g) for g in user["groups"]]
            if proj_memberships := user.get("projects"):
                proj_memberships_expanded = [regex.sub("^:", f"{project_shortname}:", p) for p in proj_memberships]
                if f"{project_shortname}:admin" in proj_memberships_expanded:
                    with contextlib.suppress(ValueError):
                        proj_memberships_expanded.remove(f"{project_shortname}:member")
                users_original[index]["projects"] = proj_memberships_expanded

        # sort both lists
        users_original = sorted(users_original or [], key=lambda x: cast(str, x["username"]))
        users_returned = sorted(users_returned or [], key=lambda x: cast(str, x["username"]))
        if len(users_original) != len(users_returned):
            self.assertEqual(
                users_original,
                users_returned,
                msg="Returned number of users is different from original number of users.",
            )
        else:
            for orig, ret in zip(users_original, users_returned):
                self.assertEqual(
                    orig,
                    ret,
                    msg=f"User with original name '{orig['username']}' and returned name '{ret['username']}' failed.",
                )

    def _compare_lists(
        self,
        lists_original: Optional[list[dict[str, Any]]],
        lists_returned: Optional[list[dict[str, Any]]],
    ) -> None:
        """
        Compare the "lists" section of the original JSON project definition
        with the "lists" section of the JSON returned by the "get" command.
        Fails with a message if the sections are not identical.

        In order to do so,
        this method removes all lists of which the nodes are defined in an Excel file,
        because they cannot be compared.

        Args:
            lists_original: "lists" section of the original file.
            lists_returned: "lists" section of the returned file.
        """
        # avoid mutable default argument
        lists_original = lists_original or []
        lists_returned = lists_returned or []

        # remove lists of which the nodes were defined in an Excel file
        for list_original in lists_original:
            if (
                isinstance(list_original["nodes"], dict)
                and len(list_original["nodes"]) == 1
                and "folder" in list_original["nodes"]
            ):
                lists_original.remove(list_original)
                lists_returned = [x for x in lists_returned if x["name"] != list_original["name"]]

        # sort both lists
        lists_original = sorted(lists_original, key=lambda x: cast(str, x.get("name", "")))
        lists_returned = sorted(lists_returned, key=lambda x: cast(str, x.get("name", "")))
        if len(lists_original) != len(lists_returned):
            self.assertEqual(
                lists_original,
                lists_returned,
                msg="Returned number of lists is different from original number of lists.",
            )
        else:
            for orig, ret in zip(lists_original, lists_returned):
                self.assertEqual(
                    orig,
                    ret,
                    msg=f"List with original name '{orig['name']}' and returned name '{ret['name']}' failed.",
                )

    def _compare_properties(
        self,
        properties_original: Optional[list[dict[str, Any]]],
        properties_returned: Optional[list[dict[str, Any]]],
        onto_name: str,
    ) -> None:
        """
        Compare the "properties" section of an onto of the original JSON project definition
        with the "properties" section of the respective onto of the JSON returned by the "get" command.
        Fails with a message if the sections are not identical.

        Args:
            properties_original: "properties" section of the original file.
            properties_returned: "properties" section of the returned file.
            onto_name: name of the ontology
        """
        # avoid mutable default argument
        properties_original = properties_original or []
        properties_returned = properties_returned or []

        # The original might have names in the form "currentonto:hasSimpleText".
        # The returned file will have ":hasSimpleText" --> remove the onto name.
        for prop in properties_original:
            if any(sup.startswith(onto_name) for sup in prop["super"]):
                prop["super"] = [regex.sub(rf"^{onto_name}:", ":", sup) for sup in prop["super"]]

        # sort both lists
        properties_original = sorted(properties_original, key=lambda x: cast(str, x.get("name", "")))
        properties_returned = sorted(properties_returned, key=lambda x: cast(str, x.get("name", "")))
        for proplist in [properties_original, properties_returned]:
            for prop in proplist:
                if isinstance(prop["super"], list):
                    prop["super"] = sorted(prop["super"])

        if len(properties_original) != len(properties_returned):
            self.assertEqual(
                properties_original,
                properties_returned,
                msg=f"Onto {onto_name}: Returned number of properties is different from original number of properties.",
            )
        else:
            for orig, ret in zip(properties_original, properties_returned):
                self.assertEqual(
                    orig,
                    ret,
                    msg=f"Onto '{onto_name}': Property with original name '{orig['name']}' "
                    f"and returned name '{ret['name']}' failed.",
                )

    def _compare_resources(
        self,
        resources_original: Optional[list[dict[str, Any]]],
        resources_returned: Optional[list[dict[str, Any]]],
        onto_name: str,
    ) -> None:
        """
        Compare the "resources" section of an onto of the original JSON project definition
        with the "resources" section of the respective onto of the JSON returned by the "get" command.
        Fails with a message if the sections are not identical.

        Args:
            resources_original: "resources" section of the original file.
            resources_returned: "resources" section of the returned file.
            onto_name: name of the ontology
        """
        # avoid mutable default argument
        resources_original = resources_original or []
        resources_returned = resources_returned or []

        resources_original = self._remove_onto_names_in_original_resources(onto_name, resources_original)

        resources_original, resources_returned = self._remove_cardinalities_in_original_resources(
            onto_name, resources_original, resources_returned
        )

        resources_original, resources_returned = self._sort_resources(resources_original, resources_returned)

        if len(resources_original) != len(resources_returned):
            self.assertEqual(
                resources_original,
                resources_returned,
                msg=f"Onto {onto_name}: Returned number of resources is different from original number of resources.",
            )
        else:
            for orig, ret in zip(resources_original, resources_returned):
                if orig.get("cardinalities") != ret.get("cardinalities"):
                    self.assertEqual(
                        orig.get("cardinalities"),
                        ret.get("cardinalities"),
                        msg=f"Onto '{onto_name}': The cardinalities of resource with original name '{orig['name']}' "
                        f"and returned name '{ret['name']}' failed.",
                    )
                else:
                    self.assertEqual(
                        orig,
                        ret,
                        msg=f"Onto '{onto_name}': Resource with original name '{orig['name']}' and returned name "
                        "'{ret['name']}' failed. The reason of the error lies OUTSIDE of the 'cardinalities' section.",
                    )

    @staticmethod
    def _sort_resources(
        resources_original: list[dict[str, Any]], resources_returned: list[dict[str, Any]]
    ) -> tuple[list[dict[str, Any]], list[dict[str, Any]]]:
        resources_original = sorted(resources_original, key=lambda x: cast(str, x.get("name", "")))
        resources_returned = sorted(resources_returned, key=lambda x: cast(str, x.get("name", "")))
        for reslist in [resources_original, resources_returned]:
            for res in reslist:
                if isinstance(res["super"], list):
                    res["super"] = sorted(res["super"])
                if res.get("cardinalities"):
                    res["cardinalities"] = sorted(res["cardinalities"], key=lambda x: cast(str, x["propname"]))
        return resources_original, resources_returned

    @staticmethod
    def _remove_cardinalities_in_original_resources(
        onto_name: str, resources_original: list[dict[str, Any]], resources_returned: list[dict[str, Any]]
    ) -> tuple[list[dict[str, Any]], list[dict[str, Any]]]:
        # If a subclass doesn't explicitly define all cardinalities of its superclass
        # (or a subproperty of a cardinality of its superclass),
        # these cardinalities are implicitly added, so the "get" command will return them.
        # It would be too complicated to test this behaviour,
        # --> remove the cardinalities of all subclasses from both original file and returned file.
        for res in resources_original:
            supers_as_list = [res["super"]] if isinstance(res["super"], str) else res["super"]
            for sup in supers_as_list:
                if regex.search(rf"^({onto_name})?:\w+$", sup):
                    if res.get("cardinalities"):
                        del res["cardinalities"]
                    # remove from returned file, too
                    res_returned = next(x for x in resources_returned if x["name"] == res["name"])
                    if res_returned.get("cardinalities"):
                        del res_returned["cardinalities"]
        return resources_original, resources_returned

    @staticmethod
    def _remove_onto_names_in_original_resources(
        onto_name: str, resources_original: list[dict[str, Any]]
    ) -> list[dict[str, Any]]:
        # The original might have names in the form "currentonto:hasSimpleText".
        # The returned file will have ":hasSimpleText" --> remove the onto name.
        for res in resources_original:
            for card in res.get("cardinalities", []):
                if card["propname"].startswith(onto_name):
                    card["propname"] = regex.sub(rf"^{onto_name}:", ":", card["propname"])
            supers_as_list = [res["super"]] if isinstance(res["super"], str) else res["super"]
            if any(sup.startswith(onto_name) for sup in supers_as_list):
                res["super"] = [regex.sub(rf"^{onto_name}:", ":", sup) for sup in supers_as_list]
        return resources_original

    @staticmethod
    def _get_most_recent_glob_match(glob_pattern: Union[str, Path]) -> Path:
        """
        Find the most recently created file that matches a glob pattern.

        Args:
            glob_pattern: glob pattern, either absolute or relative to the cwd of the caller

        Returns:
            the most recently created file that matches the glob pattern
        """
        candidates = [Path(x) for x in glob.glob(str(glob_pattern))]
        return max(candidates, key=lambda item: item.stat().st_ctime)



from __future__ import annotations

import warnings
from datetime import datetime

from loguru import logger
from lxml import etree

from dsp_tools.commands.xmlupload.models.deserialise.xmlpermission import XmlPermission
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import XMLResource
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.prepare_xml_input.check_consistency_with_ontology import (
    do_xml_consistency_check_with_ontology,
)
from dsp_tools.commands.xmlupload.prepare_xml_input.iiif_uri_validator import IIIFUriValidator
from dsp_tools.commands.xmlupload.prepare_xml_input.ontology_client import OntologyClient
from dsp_tools.commands.xmlupload.stash.stash_circular_references import identify_circular_references
from dsp_tools.commands.xmlupload.stash.stash_circular_references import stash_circular_references
from dsp_tools.commands.xmlupload.stash.stash_models import Stash
from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.models.exceptions import BaseError
from dsp_tools.models.exceptions import UserError
from dsp_tools.models.projectContext import ProjectContext
from dsp_tools.utils.connection import Connection


def prepare_upload_from_root(
    root: etree._Element,
    ontology_client: OntologyClient,
) -> tuple[list[XMLResource], dict[str, Permissions], Stash | None]:
    """Do the consistency check, resolve circular references, and return the resources and permissions."""
    do_xml_consistency_check_with_ontology(onto_client=ontology_client, root=root)
    return _resolve_circular_references(
        root=root,
        con=ontology_client.con,
        default_ontology=ontology_client.default_ontology,
    )


def _validate_iiif_uris(root: etree._Element) -> None:
    uris = [uri.strip() for node in root.iter(tag="iiif-uri") if (uri := node.text)]
    if problems := IIIFUriValidator(uris).validate():
        msg = problems.get_msg()
        warnings.warn(DspToolsUserWarning(msg))
        logger.warning(msg)


def _resolve_circular_references(
    root: etree._Element,
    con: Connection,
    default_ontology: str,
) -> tuple[list[XMLResource], dict[str, Permissions], Stash | None]:
    logger.info("Checking resources for circular references...")
    print(f"{datetime.now()}: Checking resources for circular references...")
    stash_lookup, upload_order = identify_circular_references(root)
    logger.info("Get data from XML...")
    resources, permissions_lookup = _get_data_from_xml(
        con=con,
        root=root,
        default_ontology=default_ontology,
    )
    sorting_lookup = {res.res_id: res for res in resources}
    resources = [sorting_lookup[res_id] for res_id in upload_order]
    logger.info("Stashing circular references...")
    print(f"{datetime.now()}: Stashing circular references...")
    stash = stash_circular_references(resources, stash_lookup, permissions_lookup)
    return resources, permissions_lookup, stash


def _get_data_from_xml(
    con: Connection,
    root: etree._Element,
    default_ontology: str,
) -> tuple[list[XMLResource], dict[str, Permissions]]:
    proj_context = _get_project_context_from_server(connection=con, shortcode=root.attrib["shortcode"])
    permissions = _extract_permissions_from_xml(root, proj_context)
    resources = _extract_resources_from_xml(root, default_ontology)
    permissions_lookup = {name: perm.get_permission_instance() for name, perm in permissions.items()}
    return resources, permissions_lookup


def _get_project_context_from_server(connection: Connection, shortcode: str) -> ProjectContext:
    """
    This function retrieves the project context previously uploaded on the server (json file)

    Args:
        connection: connection to the server
        shortcode: shortcode of the project

    Returns:
        Project context

    Raises:
        UserError: If the project was not previously uploaded on the server
    """
    try:
        proj_context = ProjectContext(con=connection, shortcode=shortcode)
    except BaseError:
        logger.opt(exception=True).error("Unable to retrieve project context from DSP server")
        raise UserError("Unable to retrieve project context from DSP server") from None
    return proj_context


def _extract_permissions_from_xml(root: etree._Element, proj_context: ProjectContext) -> dict[str, XmlPermission]:
    permission_ele = list(root.iter(tag="permissions"))
    permissions = [XmlPermission(permission, proj_context) for permission in permission_ele]
    return {permission.permission_id: permission for permission in permissions}


def _extract_resources_from_xml(root: etree._Element, default_ontology: str) -> list[XMLResource]:
    resources = list(root.iter(tag="resource"))
    return [XMLResource.from_node(res, default_ontology) for res in resources]



from rdflib import Graph
from rdflib import Namespace

API_SHAPES = Namespace("http://api.knora.org/ontology/knora-api/shapes/v2#")


def construct_file_value_cardinality(onto: Graph) -> Graph:
    """
    Returns cardinalities for classes with a file value.

    Args:
        onto: project ontologies

    Returns:
        Graph with file cardinalities
    """
    val_prop_mapper = {
        "ArchiveRepresentation": "hasArchiveFileValue",
        "AudioRepresentation": "hasAudioFileValue",
        "DocumentRepresentation": "hasDocumentFileValue",
        "MovingImageRepresentation": "hasMovingImageFileValue",
        "StillImageRepresentation": "hasStillImageFileValue",
        "TextRepresentation": "hasTextFileValue",
    }

    def as_class_type_and_shacl_shape(cls_name: str) -> tuple[str, str]:
        return f"knora-api:{cls_name}", f"api-shapes:{val_prop_mapper[cls_name]}_PropShape"

    g = Graph()
    for t in val_prop_mapper.keys():
        representation_type, shacl_shape = as_class_type_and_shacl_shape(t)
        g += _construct_one_representation_shape(onto, representation_type, shacl_shape)
    return g


def _construct_one_representation_shape(onto: Graph, representation_type: str, shacl_shape: str) -> Graph:
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
    PREFIX dash: <http://datashapes.org/dash#>

    CONSTRUCT {
        ?class sh:property %(shacl_shape)s .
    } WHERE {
        ?class a owl:Class ;
               rdfs:subClassOf %(representation_type)s .
    }
    """ % {"representation_type": representation_type, "shacl_shape": shacl_shape}  # noqa: UP031 (printf-string-formatting)
    if results_graph := onto.query(query_s).graph:
        return results_graph
    return Graph()



import pytest
from lxml import etree


@pytest.fixture
def resource_empty() -> etree._Element:
    return etree.fromstring("""
    <resource label="lbl" restype="http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything" id="one">
    </resource>
    """)


@pytest.fixture
def root_resource_with_props() -> etree._Element:
    return etree.fromstring("""
    <knora>
        <resource label="lbl" restype="http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything" id="one">
            <boolean-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean">
                <boolean>true</boolean>
            </boolean-prop>
            <color-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testColor">
                <color>#00ff00</color>
                <color>#00ff11</color>
            </color-prop>
        </resource>
    </knora>
    """)


@pytest.fixture
def root_resource_region() -> etree._Element:
    return etree.fromstring("""
    <knora>
        <resource restype="http://api.knora.org/ontology/knora-api/v2#Region" label="Region" id="region_1">
            <color-prop name="http://api.knora.org/ontology/knora-api/v2#hasColor">
                <color>#5d1f1e</color>
            </color-prop>
            <resptr-prop name="http://api.knora.org/ontology/knora-api/v2#isRegionOf">
                <resptr>image_thing_0</resptr>
            </resptr-prop>
            <geometry-prop name="http://api.knora.org/ontology/knora-api/v2#hasGeometry">
                <geometry>
                    {
                    "status": "active",
                    "type": "polygon",
                    "lineWidth": 5,
                    "points": [{"x": 0.4, "y": 0.6},
                    {"x": 0.5, "y": 0.9},
                    {"x": 0.8, "y": 0.9},
                    {"x": 0.7, "y": 0.6}]
                    }
                </geometry>
            </geometry-prop>
            <text-prop name="http://api.knora.org/ontology/knora-api/v2#hasComment">
                <text encoding="xml">
                    This is a polygon-formed region of interest of an image. It is also displayed as Annotation.
                </text>
            </text-prop>
        </resource>
    </knora>
    """)


@pytest.fixture
def boolean_value_corr() -> etree._Element:
    return etree.fromstring("""
        <boolean-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean">
            <boolean>true</boolean>
        </boolean-prop>
    """)


@pytest.fixture
def color_value_corr() -> etree._Element:
    return etree.fromstring("""
        <color-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testColor">
            <color>#00ff00</color>
        </color-prop>
        """)


@pytest.fixture
def color_value_corr_several() -> etree._Element:
    return etree.fromstring("""
        <color-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testColor">
            <color>#00ff00</color>
            <color>#00ff11</color>
        </color-prop>
        """)


@pytest.fixture
def date_value_corr() -> etree._Element:
    return etree.fromstring("""
        <date-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testSubDate1">
            <date>JULIAN:BCE:0700:BCE:0600</date>
        </date-prop>
        """)


@pytest.fixture
def date_value_corr_several() -> etree._Element:
    return etree.fromstring("""
        <date-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testSubDate1">
            <date>JULIAN:BCE:0700:BCE:0600</date>
            <date>ISLAMIC:BCE:0700:BCE:0600</date>
        </date-prop>
        """)


@pytest.fixture
def decimal_value_corr() -> etree._Element:
    return etree.fromstring("""
        <decimal-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testDecimalSimpleText">
            <decimal>2.71</decimal>
        </decimal-prop>
    """)


@pytest.fixture
def decimal_value_corr_several() -> etree._Element:
    return etree.fromstring("""
        <decimal-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testDecimalSimpleText">
            <decimal>1.0</decimal>
            <decimal>2.0</decimal>
        </decimal-prop>
    """)


@pytest.fixture
def geometry_value_corr() -> etree._Element:
    return etree.fromstring("""
        <geometry-prop name="hasGeometry">
            <geometry>
                {
                    "status": "active",
                    "type": "polygon",
                    "lineWidth": 5,
                    "points": [{"x": 0.4, "y": 0.6},
                               {"x": 0.5, "y": 0.9},
                               {"x": 0.8, "y": 0.9},
                               {"x": 0.7, "y": 0.6}]
                }
            </geometry>
        </geometry-prop>
    """)


@pytest.fixture
def geometry_value_wrong() -> etree._Element:
    return etree.fromstring("""
        <geometry-prop name="hasGeometry">
            <geometry></geometry>
        </geometry-prop>
    """)


@pytest.fixture
def geoname_value_corr() -> etree._Element:
    return etree.fromstring("""
        <geoname-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testGeoname">
            <geoname>1111111</geoname>
        </geoname-prop>
    """)


@pytest.fixture
def geoname_value_corr_several() -> etree._Element:
    return etree.fromstring("""
        <geoname-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testGeoname">
            <geoname>1111111</geoname>
            <geoname>2222222</geoname>
        </geoname-prop>
    """)


@pytest.fixture
def integer_value_corr() -> etree._Element:
    return etree.fromstring("""
        <integer-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testIntegerSimpleText">
            <integer>1</integer>
        </integer-prop>
    """)


@pytest.fixture
def integer_value_corr_several() -> etree._Element:
    return etree.fromstring("""
        <integer-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testIntegerSimpleText">
            <integer>1</integer>
            <integer>2</integer>
        </integer-prop>
    """)


@pytest.fixture
def list_value_corr() -> etree._Element:
    return etree.fromstring("""
        <list-prop list="firstList" name="http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp">
            <list>n1</list>
        </list-prop>
    """)


@pytest.fixture
def list_value_corr_several() -> etree._Element:
    return etree.fromstring("""
        <list-prop list="firstList" name="http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp">
            <list>n1</list>
            <list>n2</list>
        </list-prop>
    """)


@pytest.fixture
def list_value_wrong_node() -> etree._Element:
    return etree.fromstring("""
        <list-prop list="firstList" name="http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp">
            <list>other</list>
        </list-prop>
    """)


@pytest.fixture
def list_value_wrong_list() -> etree._Element:
    return etree.fromstring("""
        <list-prop list="other" name="http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp">
            <list>n1</list>
        </list-prop>
    """)


@pytest.fixture
def resptr_value_corr() -> etree._Element:
    return etree.fromstring("""
        <resptr-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo">
            <resptr>id_1</resptr>
        </resptr-prop>
    """)


@pytest.fixture
def resptr_value_corr_several() -> etree._Element:
    return etree.fromstring("""
        <resptr-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo">
            <resptr>id_1</resptr>
            <resptr>id_2</resptr>
        </resptr-prop>
    """)


@pytest.fixture
def resptr_value_wrong() -> etree._Element:
    return etree.fromstring("""
        <resptr-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo">
            <resptr></resptr>
        </resptr-prop>
    """)


@pytest.fixture
def text_richtext_value_corr() -> etree._Element:
    return etree.fromstring("""
        <text-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testRichtext">
            <text encoding="xml"><p>Text</p></text>
        </text-prop>
    """)


@pytest.fixture
def text_richtext_value_corr_several() -> etree._Element:
    return etree.fromstring("""
        <text-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testRichtext">
            <text encoding="xml">Text 1</text>
            <text encoding="xml">Text 2</text>
        </text-prop>
    """)


@pytest.fixture
def text_richtext_value_wrong() -> etree._Element:
    return etree.fromstring("""
        <text-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testRichtext">
            <text encoding="xml"></text>
        </text-prop>
    """)


@pytest.fixture
def text_simpletext_value_corr() -> etree._Element:
    return etree.fromstring("""
        <text-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testTextarea">
            <text encoding="utf8">Text</text>
        </text-prop>
    """)


@pytest.fixture
def text_simpletext_value_corr_several() -> etree._Element:
    return etree.fromstring("""
        <text-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testSimpleText">
            <text encoding="utf8">Text 1</text>
            <text encoding="utf8">Text 2</text>
        </text-prop>
    """)


@pytest.fixture
def text_simpletext_value_wrong() -> etree._Element:
    return etree.fromstring("""
        <text-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testSimpleText">
            <text encoding="utf8"></text>
        </text-prop>
    """)


@pytest.fixture
def time_value_corr() -> etree._Element:
    return etree.fromstring("""
        <time-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testTimeValue">
            <time>2019-10-23T13:45:12.01-14:00</time>
        </time-prop>
    """)


@pytest.fixture
def time_value_corr_several() -> etree._Element:
    return etree.fromstring("""
        <time-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testTimeValue">
            <time>2019-10-23T13:45:12.01-14:00</time>
            <time>2019-10-23T13:45:12.01-08:00</time>
        </time-prop>
    """)


@pytest.fixture
def uri_value_corr() -> etree._Element:
    return etree.fromstring("""
        <uri-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue">
            <uri>https://dasch.swiss</uri>
        </uri-prop>
    """)


@pytest.fixture
def uri_value_corr_several() -> etree._Element:
    return etree.fromstring("""
        <uri-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue">
            <uri>https://dasch.swiss</uri>
            <uri>https://app.dasch.swiss</uri>
        </uri-prop>
    """)


@pytest.fixture
def uri_value_wrong() -> etree._Element:
    return etree.fromstring("""
        <uri-prop name="http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue">
            <uri>oth er</uri>
        </uri-prop>
    """)



from lxml import etree

from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import XMLResource
from dsp_tools.commands.xmlupload.models.upload_state import UploadState
from dsp_tools.commands.xmlupload.upload_config import UploadConfig
from dsp_tools.commands.xmlupload.xmlupload import _tidy_up_resource_creation_idempotent


def test_idempotency_on_success() -> None:
    xml_strings = [
        """
        <resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id">
            <text-prop name=":hasSimpleText"><text encoding="utf8">foo_1 text</text></text-prop>
        </resource>
        """,
        """
        <resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id">
            <text-prop name=":hasSimpleText"><text encoding="utf8">foo_2 text</text></text-prop>
        </resource>
        """,
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "onto") for xml_str in xml_strings]
    upload_state = UploadState(xml_resources.copy(), None, UploadConfig(), {})
    for _ in range(3):
        _tidy_up_resource_creation_idempotent(upload_state, "foo_1_iri", xml_resources[0])
        assert upload_state.pending_resources == xml_resources[1:]
        assert upload_state.failed_uploads == []
        assert upload_state.iri_resolver.lookup == {"foo_1_id": "foo_1_iri"}
        assert not upload_state.pending_stash


def test_idempotency_on_failure() -> None:
    xml_strings = [
        """
        <resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id">
            <text-prop name=":hasSimpleText"><text encoding="utf8">foo_1 text</text></text-prop>
        </resource>
        """,
        """
        <resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id">
            <text-prop name=":hasSimpleText"><text encoding="utf8">foo_2 text</text></text-prop>
        </resource>
        """,
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "onto") for xml_str in xml_strings]
    upload_state = UploadState(xml_resources.copy(), None, UploadConfig(), {})
    for _ in range(3):
        _tidy_up_resource_creation_idempotent(upload_state, None, xml_resources[0])
        assert upload_state.pending_resources == xml_resources[1:]
        assert upload_state.failed_uploads == ["foo_1_id"]
        assert upload_state.iri_resolver.lookup == {}
        assert not upload_state.pending_stash



from __future__ import annotations

import warnings
from copy import deepcopy
from dataclasses import dataclass
from typing import Any
from typing import Protocol

from lxml import etree
from namedentities.core import numeric_entities  # type: ignore[import-untyped]

from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.models.exceptions import InputError
from dsp_tools.utils.uri_util import is_uri
from dsp_tools.xmllib.helpers import escape_reserved_xml_characters
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.value_checkers import check_richtext_syntax
from dsp_tools.xmllib.value_checkers import is_color
from dsp_tools.xmllib.value_checkers import is_date
from dsp_tools.xmllib.value_checkers import is_decimal
from dsp_tools.xmllib.value_checkers import is_geoname
from dsp_tools.xmllib.value_checkers import is_integer
from dsp_tools.xmllib.value_checkers import is_string_like
from dsp_tools.xmllib.value_checkers import is_timestamp
from dsp_tools.xmllib.value_converters import convert_to_bool

XML_NAMESPACE_MAP = {None: "https://dasch.swiss/schema", "xsi": "http://www.w3.org/2001/XMLSchema-instance"}
DASCH_SCHEMA = "{https://dasch.swiss/schema}"


class Value(Protocol):
    value: Any
    prop_name: str
    permissions: Permissions
    comment: str | None
    resource_id: str | None

    def serialise(self) -> etree._Element:
        raise NotImplementedError

    def make_prop(self) -> etree._Element:
        raise NotImplementedError

    def make_element(self) -> etree._Element:
        raise NotImplementedError


@dataclass
class BooleanValue(Value):
    value: Any
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        try:
            converted_bool = convert_to_bool(self.value)
            self.value = str(converted_bool).lower()
        except InputError:
            _warn_type_mismatch(
                expected_type="bool", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )
            self.value = str(self.value)

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}boolean-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {}
        if self.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
            attribs["permissions"] = self.permissions.value
        if is_string_like(self.comment):
            attribs["comment"] = str(self.comment)
        ele = etree.Element(f"{DASCH_SCHEMA}boolean", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class ColorValue(Value):
    value: Any
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_color(self.value):
            _warn_type_mismatch(
                expected_type="color", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}color-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {}
        if self.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
            attribs["permissions"] = self.permissions.value
        if is_string_like(self.comment):
            attribs["comment"] = str(self.comment)
        ele = etree.Element(f"{DASCH_SCHEMA}color", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class DateValue(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_date(self.value):
            _warn_type_mismatch(
                expected_type="date", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}date-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {}
        if self.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
            attribs["permissions"] = self.permissions.value
        if is_string_like(self.comment):
            attribs["comment"] = str(self.comment)
        ele = etree.Element(f"{DASCH_SCHEMA}date", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class DecimalValue(Value):
    value: Any
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_decimal(self.value):
            _warn_type_mismatch(
                expected_type="decimal", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}decimal-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {}
        if self.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
            attribs["permissions"] = self.permissions.value
        if is_string_like(self.comment):
            attribs["comment"] = str(self.comment)
        ele = etree.Element(f"{DASCH_SCHEMA}decimal", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class GeonameValue(Value):
    value: Any
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_geoname(self.value):
            _warn_type_mismatch(
                expected_type="geoname", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}geoname-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {}
        if self.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
            attribs["permissions"] = self.permissions.value
        if is_string_like(self.comment):
            attribs["comment"] = str(self.comment)
        ele = etree.Element(f"{DASCH_SCHEMA}geoname", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class IntValue(Value):
    value: int | str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_integer(self.value):
            _warn_type_mismatch(
                expected_type="integer", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}integer-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {}
        if self.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
            attribs["permissions"] = self.permissions.value
        if is_string_like(self.comment):
            attribs["comment"] = str(self.comment)
        ele = etree.Element(f"{DASCH_SCHEMA}integer", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class LinkValue(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_string_like(self.value):
            _warn_type_mismatch(
                expected_type="string", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}resptr-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {}
        if self.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
            attribs["permissions"] = self.permissions.value
        if is_string_like(self.comment):
            attribs["comment"] = str(self.comment)
        ele = etree.Element(f"{DASCH_SCHEMA}resptr", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class ListValue(Value):
    value: Any
    list_name: Any
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_string_like(self.value) or not is_string_like(self.list_name):
            _warn_type_mismatch(
                expected_type="list", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(
            f"{DASCH_SCHEMA}list-prop", name=self.prop_name, list=self.list_name, nsmap=XML_NAMESPACE_MAP
        )

    def make_element(self) -> etree._Element:
        attribs = {}
        if self.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
            attribs["permissions"] = self.permissions.value
        if is_string_like(self.comment):
            attribs["comment"] = str(self.comment)
        ele = etree.Element(f"{DASCH_SCHEMA}list", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class SimpleText(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_string_like(self.value):
            _warn_type_mismatch(
                expected_type="string", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}text-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {"encoding": "utf8"}
        if self.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
            attribs["permissions"] = self.permissions.value
        if is_string_like(self.comment):
            attribs["comment"] = str(self.comment)
        ele = etree.Element(f"{DASCH_SCHEMA}text", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class Richtext(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_string_like(self.value):
            _warn_type_mismatch(
                expected_type="string", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )
        check_richtext_syntax(self.value)

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}text-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {"encoding": "xml"}
        if self.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
            attribs["permissions"] = self.permissions.value
        if is_string_like(self.comment):
            attribs["comment"] = str(self.comment)
        ele = etree.Element(f"{DASCH_SCHEMA}text", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        return self._create_richtext_elements_from_string(ele)

    def _create_richtext_elements_from_string(self, text_element: etree._Element) -> etree._Element:
        new_element = deepcopy(text_element)
        escaped_text = escape_reserved_xml_characters(self.value)
        # transform named entities (=character references) to numeric entities, e.g. &nbsp; -> &#160;
        num_ent = numeric_entities(escaped_text)
        pseudo_xml = f"<ignore-this>{num_ent}</ignore-this>"
        try:
            parsed = etree.fromstring(pseudo_xml)
        except etree.XMLSyntaxError as err:
            msg = (
                f"The resource with the ID '{self.resource_id}' and the property '{self.prop_name}' "
                f"contains richtext that is invalid XMl."
            )
            msg += f"\nOriginal error message: {err.msg}"
            msg += f"\nEventual line/column numbers are relative to this text: {pseudo_xml}"
            raise InputError(msg) from None
        new_element.text = parsed.text  # everything before the first child tag
        new_element.extend(list(parsed))  # all (nested) children of the pseudo-xml
        return new_element


@dataclass
class TimeValue(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_timestamp(self.value):
            _warn_type_mismatch(
                expected_type="timestamp", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}time-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {}
        if self.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
            attribs["permissions"] = self.permissions.value
        if is_string_like(self.comment):
            attribs["comment"] = str(self.comment)
        ele = etree.Element(f"{DASCH_SCHEMA}time", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class UriValue(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_uri(self.value):
            _warn_type_mismatch(
                expected_type="uri", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}uri-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {}
        if self.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
            attribs["permissions"] = self.permissions.value
        if is_string_like(self.comment):
            attribs["comment"] = str(self.comment)
        ele = etree.Element(f"{DASCH_SCHEMA}uri", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


def _warn_type_mismatch(expected_type: str, value: Any, prop_name: str, res_id: str | None) -> None:
    """Emits a warning if a values is not in the expected format."""
    msg = f"At the following location a '{expected_type}' does not conform to the expected format.\n"
    msg += f"Resource: {res_id} | " if res_id else ""
    msg += f"Value: {value} | Property: {prop_name}"
    warnings.warn(DspToolsUserWarning(msg))



from __future__ import annotations

import json
from dataclasses import dataclass
from json.decoder import JSONDecodeError
from typing import Callable
from typing import TypeAlias
from typing import Union
from typing import assert_never

from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryValue
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryValueTypes
from dsp_tools.commands.xmlupload.models.intermediary.values import IntervalFloats
from dsp_tools.models.exceptions import InputError
from dsp_tools.utils.date_util import Date
from dsp_tools.utils.date_util import parse_date_string

InputTypes: TypeAlias = Union[str, FormattedTextValue]


@dataclass
class TypeTransformerMapper:
    val_type: type[IntermediaryValue]
    val_transformer: Callable[[InputTypes], IntermediaryValueTypes]


def transform_boolean(value: InputTypes) -> bool:
    """Transform the value into a boolean"""
    match value:
        case "True" | "true" | "1" | 1 | True:
            return True
        case "False" | "false" | "0" | 0 | False:
            return False
        case _:
            raise InputError(f"Could not parse boolean value: {value}")


def transform_date(input_value: InputTypes) -> Date:
    """Transform a date string into a date object."""
    val = assert_is_string(input_value)
    return parse_date_string(val)


def transform_decimal(value: InputTypes) -> float:
    """Transform a value into a float"""
    str_val = assert_is_string(value)
    return float(str_val)


def transform_integer(value: InputTypes) -> int:
    """Transform a value into an integer"""
    str_val = assert_is_string(value)
    return int(str_val)


def transform_interval(input_value: InputTypes) -> IntervalFloats:
    """Transform a sting input into an interval object."""
    val = assert_is_string(input_value)
    split_val = [res for x in val.split(":", 1) if (res := x.strip())]
    if not len(split_val) == 2:
        raise InputError(f"Could not parse interval: {val}")
    try:
        return IntervalFloats(float(split_val[0]), float(split_val[1]))
    except ValueError:
        raise InputError(f"Could not parse interval: {val}") from None


def transform_geometry(value: InputTypes) -> str:
    """Transform a value into a geometry string"""
    str_val = assert_is_string(value)
    try:
        return json.dumps(json.loads(str_val))
    except JSONDecodeError:
        raise InputError(f"Could not parse json value: {value}") from None


def assert_is_string(value: str | FormattedTextValue) -> str:
    """Assert a value is a string."""
    match value:
        case str() as s:
            return s
        case FormattedTextValue() as xml:
            raise InputError(f"Expected string value, but got XML value: {xml.as_xml()}")
        case _:
            assert_never(value)



from __future__ import annotations

from dataclasses import dataclass

from dsp_tools.commands.xmlupload.models.intermediary.file_values import IntermediaryFileValue
from dsp_tools.commands.xmlupload.models.intermediary.file_values import IntermediaryIIIFUri
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryValue
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.models.datetimestamp import DateTimeStamp


@dataclass
class ResourceTransformationOutput:
    resource_success: IntermediaryResource | None
    resource_failure: ResourceInputConversionFailure | None = None


@dataclass
class ResourceInputConversionFailure:
    resource_id: str
    failure_msg: str


@dataclass
class IntermediaryResource:
    res_id: str
    type_iri: str
    label: str
    permissions: Permissions | None
    values: list[IntermediaryValue]
    file_value: IntermediaryFileValue | None
    iiif_uri: IntermediaryIIIFUri | None
    migration_metadata: MigrationMetadata | None


@dataclass
class MigrationMetadata:
    iri_str: str | None
    creation_date: DateTimeStamp | None



import pytest
from rdflib import SH
from rdflib import Graph

from dsp_tools.commands.validate_data.sparql.file_value_shacl import construct_file_value_cardinality
from test.unittests.commands.validate_data.constants import API_SHAPES
from test.unittests.commands.validate_data.constants import ONTO


@pytest.fixture
def file_value_shacl(onto_graph: Graph) -> Graph:
    return construct_file_value_cardinality(onto_graph)


def test_construct_file_value_cardinality(file_value_shacl: Graph) -> None:
    number_of_classes_implemented = 6
    assert len(list(file_value_shacl.subjects(SH.property))) == number_of_classes_implemented


def test_construct_audio(file_value_shacl: Graph) -> None:
    result_list = list(file_value_shacl.subjects(SH.property, API_SHAPES.hasAudioFileValue_PropShape))
    assert len(result_list) == 1
    moving_image = result_list[0]
    assert moving_image == ONTO.TestAudioRepresentation


def test_construct_moving_image(file_value_shacl: Graph) -> None:
    result_list = list(file_value_shacl.subjects(SH.property, API_SHAPES.hasMovingImageFileValue_PropShape))
    assert len(result_list) == 1
    moving_image = result_list[0]
    assert moving_image == ONTO.TestMovingImageRepresentation


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from pathlib import Path
from typing import Iterable

from dsp_tools.commands.ingest_xmlupload.upload_files.input_error import FileProblems

SUPPORTED_EXTENSIONS = (
    "zip,tar,gz,z,tgz,gzip,7z,mp3,wav,pdf,doc,docx,xls,xlsx,ppt,pptx,"
    "mp4,jpg,jpeg,jp2,png,tif,tiff,odd,rng,txt,xml,xsd,xsl,csv"
).split(",")


def check_files(files: Iterable[Path]) -> FileProblems | None:
    """Check if the files exist and have supported extensions."""
    unsupported_files = [file for file in files if file.suffix[1:].casefold() not in SUPPORTED_EXTENSIONS]
    non_existing_files = [file for file in files if not file.exists() and file not in unsupported_files]
    if non_existing_files or unsupported_files:
        return FileProblems(non_existing_files, unsupported_files)
    return None



import json
import os
import time
from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from functools import partial
from importlib.metadata import version
from typing import Any
from typing import Literal
from typing import Never
from typing import cast

import regex
from loguru import logger
from requests import JSONDecodeError
from requests import ReadTimeout
from requests import RequestException
from requests import Response
from requests import Session

from dsp_tools.models.exceptions import InvalidInputError
from dsp_tools.models.exceptions import PermanentConnectionError
from dsp_tools.models.exceptions import PermanentTimeOutError
from dsp_tools.utils.authentication_client import AuthenticationClient
from dsp_tools.utils.connection import Connection
from dsp_tools.utils.logger_config import WARNINGS_SAVEPATH
from dsp_tools.utils.set_encoder import SetEncoder

HTTP_OK = 200
HTTP_UNAUTHORIZED = 401


@dataclass
class RequestParameters:
    method: Literal["POST", "GET", "PUT", "DELETE"]
    url: str
    timeout: int
    data: dict[str, Any] | None = None
    data_serialized: bytes | None = field(init=False, default=None)
    headers: dict[str, str] | None = None
    files: dict[str, tuple[str, Any]] | None = None

    def __post_init__(self) -> None:
        self.data_serialized = self._serialize_payload(self.data)

    def _serialize_payload(self, payload: dict[str, Any] | None) -> bytes | None:
        # If data is not encoded as bytes, issues can occur with non-ASCII characters,
        # where the content-length of the request will turn out to be different from the actual length.
        return json.dumps(payload, cls=SetEncoder, ensure_ascii=False).encode("utf-8") if payload else None

    def as_kwargs(self) -> dict[str, Any]:
        return {
            "method": self.method,
            "url": self.url,
            "timeout": self.timeout,
            "data": self.data_serialized,
            "headers": self.headers,
            "files": self.files,
        }


@dataclass
class ConnectionLive(Connection):
    """
    A Connection instance represents a connection to a DSP server.

    Attributes:
        server: address of the server, e.g https://api.dasch.swiss
        token: session token received by the server after login
    """

    server: str
    authenticationClient: AuthenticationClient | None = None
    session: Session = field(init=False, default=Session())
    # downtimes of server-side services -> API still processes request
    # -> retry too early has side effects (e.g. duplicated resources)
    timeout_put_post: int = field(init=False, default=30 * 60)
    timeout_get: int = field(init=False, default=20)

    def __post_init__(self) -> None:
        self.session.headers["User-Agent"] = f'DSP-TOOLS/{version("dsp-tools")}'
        if self.server.endswith("/"):
            self.server = self.server[:-1]
        if self.authenticationClient and (token := self.authenticationClient.get_token()):
            self.session.headers["Authorization"] = f"Bearer {token}"

    def logout(self) -> None:
        """
        Remove the authorization header from the connection's session.
        """
        del self.session.headers["Authorization"]

    def post(
        self,
        route: str,
        data: dict[str, Any] | None = None,
        files: dict[str, tuple[str, Any]] | None = None,
        headers: dict[str, str] | None = None,
        timeout: int | None = None,
    ) -> dict[str, Any]:
        """
        Make an HTTP POST request to the server to which this connection has been established.

        Args:
            route: route that will be called on the server
            data: payload of the HTTP request
            files: files to be uploaded, if any
            headers: headers for the HTTP request
            timeout: timeout of the HTTP request, or None if the default should be used

        Returns:
            response from server

        Raises:
            PermanentConnectionError: if all attempts have failed
            InvalidInputError: if the API responds with a permanent error because of invalid input data
        """
        if data:
            headers = headers or {}
            if "Content-Type" not in headers:
                headers["Content-Type"] = "application/json; charset=UTF-8"
        params = RequestParameters(
            "POST", self._make_url(route), timeout or self.timeout_put_post, data, headers, files
        )
        response = self._try_network_action(params)
        return cast(dict[str, Any], response.json())

    def get(
        self,
        route: str,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        """
        Make an HTTP GET request to the server to which this connection has been established.

        Args:
            route: route that will be called on the server
            headers: headers for the HTTP request

        Returns:
            response from server

        Raises:
            PermanentConnectionError: if all attempts have failed
            InvalidInputError: if the API responds with a permanent error because of invalid input data
        """
        params = RequestParameters("GET", self._make_url(route), self.timeout_get, headers=headers)
        response = self._try_network_action(params)
        return cast(dict[str, Any], response.json())

    def put(
        self,
        route: str,
        data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        """
        Make an HTTP GET request to the server to which this connection has been established.

        Args:
            route: route that will be called on the server
            data: payload of the HTTP request
            headers: headers of the HTTP request

        Returns:
            response from server

        Raises:
            PermanentConnectionError: if all attempts have failed
            InvalidInputError: if the API responds with a permanent error because of invalid input data
        """
        if data:
            headers = headers or {}
            if "Content-Type" not in headers:
                headers["Content-Type"] = "application/json; charset=UTF-8"
        params = RequestParameters("PUT", self._make_url(route), self.timeout_put_post, data, headers)
        response = self._try_network_action(params)
        return cast(dict[str, Any], response.json())

    def _make_url(self, route: str) -> str:
        if not route.startswith("/"):
            route = f"/{route}"
        return self.server + route

    def _try_network_action(self, params: RequestParameters) -> Response:
        """
        Try 7 times to execute an HTTP request.
        If a timeout error, a ConnectionError, or a requests.RequestException occur,
        or if the response indicates that there is a non-permanent server-side problem,
        this function waits and retries the HTTP request.
        The waiting times are 1, 2, 4, 8, 16, 32, 64 seconds.

        Args:
            params: keyword arguments for the HTTP request

        Raises:
            BadCredentialsError: if the server returns a 401 status code on the route /v2/authentication
            PermanentConnectionError: if all attempts have failed
            InvalidInputError: if the API responds with a permanent error because of invalid input data
            unexpected exceptions: if the action fails with an unexpected exception

        Returns:
            the return value of action
        """
        action = partial(self.session.request, **params.as_kwargs())
        for i in range(7):
            try:
                self._log_request(params)
                response = action()
            except (TimeoutError, ReadTimeout) as err:
                self._log_and_raise_timeouts(err)
            except (ConnectionError, RequestException):
                self._renew_session()
                self._log_and_sleep(reason="Connection Error raised", retry_counter=i, exc_info=True)
                continue

            self._log_response(response)
            if response.status_code == HTTP_OK:
                return response

            self._handle_non_ok_responses(response, i)

        # if all attempts have failed, raise error
        msg = f"Permanently unable to execute the network action. See {WARNINGS_SAVEPATH} for more information."
        raise PermanentConnectionError(msg)

    def _handle_non_ok_responses(self, response: Response, retry_counter: int) -> None:
        if _should_retry(response):
            self._log_and_sleep("Transient Error", retry_counter, exc_info=False)
            return None
        else:
            msg = "Permanently unable to execute the network action. "
            if original_str := regex.search(r'{"knora-api:error":"dsp\.errors\.(.*)","@context', str(response.content)):
                msg += f"\n{' '*37}Original Message: {original_str.group(1)}\n"
                if original_str.group(1).startswith("OntologyConstraintException"):
                    msg += f"See {WARNINGS_SAVEPATH} for more information."
                    raise InvalidInputError(msg)
            msg += f"See {WARNINGS_SAVEPATH} for more information."
            raise PermanentConnectionError(msg)

    def _renew_session(self) -> None:
        self.session.close()
        self.session = Session()
        self.session.headers["User-Agent"] = f'DSP-TOOLS/{version("dsp-tools")}'
        if self.authenticationClient and (token := self.authenticationClient.get_token()):
            self.session.headers["Authorization"] = f"Bearer {token}"

    def _log_and_sleep(self, reason: str, retry_counter: int, exc_info: bool) -> None:
        msg = f"{reason}: Try reconnecting to DSP server, next attempt in {2 ** retry_counter} seconds..."
        print(f"{datetime.now()}: {msg}")
        if exc_info:
            logger.opt(exception=True).error(f"{msg} ({retry_counter=:})")
        else:
            logger.error(f"{msg} ({retry_counter=:})")
        time.sleep(2**retry_counter)

    def _log_and_raise_timeouts(self, error: TimeoutError | ReadTimeout) -> Never:
        msg = f"A '{error.__class__.__name__}' occurred during the connection to the DSP server."
        print(f"{datetime.now()}: {msg}")
        logger.exception(msg)
        raise PermanentTimeOutError(msg) from None

    def _log_response(self, response: Response) -> None:
        dumpobj: dict[str, Any] = {
            "status_code": response.status_code,
            "headers": _sanitize_headers(dict(response.headers)),
        }
        try:
            dumpobj["content"] = response.json()
        except JSONDecodeError:
            dumpobj["content"] = response.text
        logger.debug(f"RESPONSE: {json.dumps(dumpobj)}")

    def _log_request(self, params: RequestParameters) -> None:
        dumpobj = {
            "method": params.method,
            "url": params.url,
            "headers": _sanitize_headers(dict(self.session.headers) | (params.headers or {})),  # type: ignore[operator]
            "timeout": params.timeout,
        }
        if params.data:
            data = params.data.copy()
            if "password" in data:
                data["password"] = "***"
            dumpobj["data"] = data
        if params.files:
            dumpobj["files"] = params.files["file"][0]
        logger.debug(f"REQUEST: {json.dumps(dumpobj, cls=SetEncoder)}")


def _sanitize_headers(headers: dict[str, str | bytes]) -> dict[str, str]:
    def _mask(key: str, value: str | bytes) -> str:
        if isinstance(value, bytes):
            value = value.decode("utf-8")
        if key == "Authorization" and value.startswith("Bearer "):
            return "Bearer ***"
        if key == "Set-Cookie":
            return "***"
        return value

    return {k: _mask(k, v) for k, v in headers.items()}


def _should_retry(response: Response) -> bool:
    in_500_range = 500 <= response.status_code < 600
    try_again_later = "try again later" in response.text.lower()
    in_testing_env = os.getenv("DSP_TOOLS_TESTING") == "true"  # set in .github/workflows/tests-on-push.yml
    return (try_again_later or in_500_range) and not in_testing_env



import pytest

from dsp_tools.commands.project.models.ontology import Ontology
from dsp_tools.commands.project.models.propertyclass import PropertyClass
from dsp_tools.models.langstring import LangString
from dsp_tools.models.langstring import Languages
from dsp_tools.utils.authentication_client_live import AuthenticationClientLive
from dsp_tools.utils.connection_live import ConnectionLive


def test_PropertyClass_create() -> None:
    auth = AuthenticationClientLive("http://0.0.0.0:3333", "root@example.com", "test")
    con = ConnectionLive("http://0.0.0.0:3333", auth)

    # Create a test ontology
    onto = Ontology(
        con=con,
        project="http://rdfh.ch/projects/0001",
        name="onto-1",
        label="onto-label",
    ).create()
    assert onto.iri is not None
    last_modification_date = onto.lastModificationDate

    # create new property class
    prop_name = "MyPropClassName"
    prop_label = LangString({Languages.DE: "MyPropClassLabel"})
    prop_comment = LangString({Languages.DE: "This is a property class for testing"})
    last_modification_date, property_class = PropertyClass(
        con=con,
        context=onto.context,
        name=prop_name,
        ontology_id=onto.iri,
        rdf_object="TextValue",
        label=prop_label,
        comment=prop_comment,
    ).create(last_modification_date)

    assert property_class.iri is not None
    assert property_class.name == prop_name
    assert property_class.label["de"] == prop_label["de"]
    assert property_class.comment["de"] == prop_comment["de"]


if __name__ == "__main__":
    pytest.main([__file__])



import pytest
from rdflib import XSD
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.commands.validate_data.make_data_rdf import _get_file_extension
from dsp_tools.commands.validate_data.make_data_rdf import _transform_file_value
from dsp_tools.commands.validate_data.make_data_rdf import _transform_one_resource
from dsp_tools.commands.validate_data.make_data_rdf import _transform_one_value
from dsp_tools.commands.validate_data.models.data_deserialised import BitstreamDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import BooleanValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import ColorValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import DateValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import DecimalValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import GeonameValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import IIIFUriDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import IntValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import LinkValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import ListValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import ResourceDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import RichtextDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import SimpleTextDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import TimeValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import UriValueDeserialised
from dsp_tools.commands.validate_data.models.data_rdf import BooleanValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import ColorValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import DateValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import DecimalValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import FileValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import GeonameValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import IntValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import LinkValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import ListValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import ResourceRDF
from dsp_tools.commands.validate_data.models.data_rdf import RichtextRDF
from dsp_tools.commands.validate_data.models.data_rdf import SimpleTextRDF
from dsp_tools.commands.validate_data.models.data_rdf import TimeValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import UriValueRDF
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import KNORA_API
from test.unittests.commands.validate_data.constants import DATA


class TestResource:
    def test_empty(self, resource_deserialised_no_values: ResourceDeserialised) -> None:
        res_li = _transform_one_resource(resource_deserialised_no_values)
        assert len(res_li) == 1
        res = res_li[0]
        assert isinstance(res, ResourceRDF)
        assert res.res_iri == DATA["id"]
        assert res.res_class == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything")
        assert res.label == Literal("lbl", datatype=XSD.string)

    def test_with_props(self, resource_deserialised_with_values: ResourceDeserialised) -> None:
        res_list = _transform_one_resource(resource_deserialised_with_values)
        assert len(res_list) == 2
        res = res_list[0]
        assert isinstance(res, ResourceRDF)
        assert res.res_iri == DATA["id"]
        assert res.res_class == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything")
        assert res.label == Literal("lbl", datatype=XSD.string)
        val = res_list[1]
        assert isinstance(val, BooleanValueRDF)


class TestBooleanValue:
    def test_corr(self, boolean_value_deserialised_corr: BooleanValueDeserialised) -> None:
        val = _transform_one_value(boolean_value_deserialised_corr, DATA["id"])
        assert isinstance(val, BooleanValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean")
        assert val.object_value == Literal("false", datatype=XSD.boolean)

    def test_one(self, boolean_value_deserialised_one: BooleanValueDeserialised) -> None:
        val = _transform_one_value(boolean_value_deserialised_one, DATA["id"])
        assert isinstance(val, BooleanValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean")
        assert val.object_value == Literal("true", datatype=XSD.boolean)

    def test_zero(self, boolean_value_deserialised_zero: BooleanValueDeserialised) -> None:
        val = _transform_one_value(boolean_value_deserialised_zero, DATA["id"])
        assert isinstance(val, BooleanValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean")
        assert val.object_value == Literal("false", datatype=XSD.boolean)

    def test_none(self, boolean_value_deserialised_none: BooleanValueDeserialised) -> None:
        val = _transform_one_value(boolean_value_deserialised_none, DATA["id"])
        assert isinstance(val, BooleanValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean")
        assert val.object_value == Literal("", datatype=XSD.string)


class TestColorValue:
    def test_corr(self, color_value_deserialised_corr: ColorValueDeserialised) -> None:
        val = _transform_one_value(color_value_deserialised_corr, DATA["id"])
        assert isinstance(val, ColorValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testColor")
        assert val.object_value == Literal("#00ff00", datatype=XSD.string)

    def test_none(self, color_value_deserialised_none: ColorValueDeserialised) -> None:
        val = _transform_one_value(color_value_deserialised_none, DATA["id"])
        assert isinstance(val, ColorValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testColor")
        assert val.object_value == Literal("", datatype=XSD.string)


class TestDateValue:
    def test_corr(self, date_value_deserialised_corr: DateValueDeserialised) -> None:
        val = _transform_one_value(date_value_deserialised_corr, DATA["id"])
        assert isinstance(val, DateValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testSubDate1")
        assert val.object_value == Literal("JULIAN:BCE:0700:BCE:0600", datatype=XSD.string)

    def test_none(self, date_value_deserialised_none: DateValueDeserialised) -> None:
        val = _transform_one_value(date_value_deserialised_none, DATA["id"])
        assert isinstance(val, DateValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testSubDate1")
        assert val.object_value == Literal("", datatype=XSD.string)


class TestDecimalValue:
    def test_corr(self, decimal_value_deserialised_corr: DecimalValueDeserialised) -> None:
        val = _transform_one_value(decimal_value_deserialised_corr, DATA["id"])
        assert isinstance(val, DecimalValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testDecimalSimpleText")
        assert val.object_value == Literal("1.2", datatype=XSD.decimal)

    def test_none(self, decimal_value_deserialised_none: DecimalValueDeserialised) -> None:
        val = _transform_one_value(decimal_value_deserialised_none, DATA["id"])
        assert isinstance(val, DecimalValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testDecimalSimpleText")
        assert val.object_value == Literal("", datatype=XSD.string)


class TestGeonameValue:
    def test_corr(self, geoname_value_deserialised_corr: GeonameValueDeserialised) -> None:
        val = _transform_one_value(geoname_value_deserialised_corr, DATA["id"])
        assert isinstance(val, GeonameValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testGeoname")
        assert val.object_value == Literal("1241345", datatype=XSD.string)

    def test_none(self, geoname_value_deserialised_none: GeonameValueDeserialised) -> None:
        val = _transform_one_value(geoname_value_deserialised_none, DATA["id"])
        assert isinstance(val, GeonameValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testGeoname")
        assert val.object_value == Literal("", datatype=XSD.string)


class TestIntValue:
    def test_corr(self, int_value_deserialised_corr: IntValueDeserialised) -> None:
        val = _transform_one_value(int_value_deserialised_corr, DATA["id"])
        assert isinstance(val, IntValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testIntegerSimpleText")
        assert val.object_value == Literal("1", datatype=XSD.integer)

    def test_none(self, int_value_deserialised_none: IntValueDeserialised) -> None:
        val = _transform_one_value(int_value_deserialised_none, DATA["id"])
        assert isinstance(val, IntValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testIntegerSimpleText")
        assert val.object_value == Literal("", datatype=XSD.string)


class TestLinkValue:
    def test_corr(self, link_value_deserialised_corr: LinkValueDeserialised) -> None:
        val = _transform_one_value(link_value_deserialised_corr, DATA["id"])
        assert isinstance(val, LinkValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo")
        assert val.object_value == DATA["link-id"]

    def test_none(self, link_value_deserialised_none: LinkValueDeserialised) -> None:
        val = _transform_one_value(link_value_deserialised_none, DATA["id"])
        assert isinstance(val, LinkValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo")
        assert val.object_value == DATA[""]


class TestListValue:
    def test_corr(self, list_value_deserialised_corr: ListValueDeserialised) -> None:
        val = _transform_one_value(list_value_deserialised_corr, DATA["id"])
        assert isinstance(val, ListValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp")
        assert val.list_name == Literal("firstList", datatype=XSD.string)
        assert val.object_value == Literal("n1", datatype=XSD.string)

    def test_none(self, list_value_deserialised_none: ListValueDeserialised) -> None:
        val = _transform_one_value(list_value_deserialised_none, DATA["id"])
        assert isinstance(val, ListValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp")
        assert val.list_name == Literal("firstList", datatype=XSD.string)
        assert val.object_value == Literal("", datatype=XSD.string)


class TestSimpleTextValue:
    def test_corr(self, simple_text_deserialised_corr: SimpleTextDeserialised) -> None:
        val = _transform_one_value(simple_text_deserialised_corr, DATA["id"])
        assert isinstance(val, SimpleTextRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testTextarea")
        assert val.object_value == Literal("simple text", datatype=XSD.string)

    def test_none(self, simple_text_deserialised_none: SimpleTextDeserialised) -> None:
        val = _transform_one_value(simple_text_deserialised_none, DATA["id"])
        assert isinstance(val, SimpleTextRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testTextarea")
        assert val.object_value == Literal("", datatype=XSD.string)


class TestRichtextValue:
    def test_corr(self, richtext_deserialised_corr: RichtextDeserialised) -> None:
        val = _transform_one_value(richtext_deserialised_corr, DATA["id"])
        assert isinstance(val, RichtextRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testRichtext")
        assert val.object_value == Literal("rich text", datatype=XSD.string)

    def test_none(self, richtext_deserialised_none: RichtextDeserialised) -> None:
        val = _transform_one_value(richtext_deserialised_none, DATA["id"])
        assert isinstance(val, RichtextRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testRichtext")
        assert val.object_value == Literal("", datatype=XSD.string)


class TestTimeValue:
    def test_corr(self, time_value_deserialised_corr: TimeValueDeserialised) -> None:
        val = _transform_one_value(time_value_deserialised_corr, DATA["id"])
        assert isinstance(val, TimeValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testTimeValue")
        assert val.object_value == Literal("2019-10-23T13:45:12.01-14:00", datatype=XSD.dateTimeStamp)

    def test_none(self, time_value_deserialised_none: TimeValueDeserialised) -> None:
        val = _transform_one_value(time_value_deserialised_none, DATA["id"])
        assert isinstance(val, TimeValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testTimeValue")
        assert val.object_value == Literal("", datatype=XSD.string)


class TestUriValue:
    def test_corr(self, uri_value_deserialised_corr: UriValueDeserialised) -> None:
        val = _transform_one_value(uri_value_deserialised_corr, DATA["id"])
        assert isinstance(val, UriValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue")
        assert val.object_value == Literal("https://dasch.swiss", datatype=XSD.anyURI)

    def test_none(self, uri_value_deserialised_none: UriValueDeserialised) -> None:
        val = _transform_one_value(uri_value_deserialised_none, DATA["id"])
        assert isinstance(val, UriValueRDF)
        assert val.res_iri == DATA["id"]
        assert val.prop_name == URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue")
        assert val.object_value == Literal("", datatype=XSD.string)


class TestTransformFileValue:
    @pytest.mark.parametrize("extension", ["zip", "tar", "gz", "z", "tgz", "gzip", "7z"])
    def test_archive_file(self, extension: str) -> None:
        bitstream = BitstreamDeserialised("id", f"test.{extension}")
        result = _transform_file_value(bitstream)
        assert isinstance(result, FileValueRDF)
        assert result.prop_type_info.knora_type == KNORA_API.ArchiveFileValue
        assert result.value == Literal(bitstream.value, datatype=XSD.string)
        bitstream = BitstreamDeserialised("id", f"test.{extension.upper()}")
        result = _transform_file_value(bitstream)
        assert isinstance(result, FileValueRDF)
        assert result.prop_type_info.knora_type == KNORA_API.ArchiveFileValue
        assert result.value == Literal(bitstream.value, datatype=XSD.string)

    @pytest.mark.parametrize("extension", ["mp3", "wav"])
    def test_audio_file(self, extension: str) -> None:
        bitstream = BitstreamDeserialised("id", f"test.{extension}")
        result = _transform_file_value(bitstream)
        assert isinstance(result, FileValueRDF)
        assert result.prop_type_info.knora_type == KNORA_API.AudioFileValue
        assert result.value == Literal(bitstream.value, datatype=XSD.string)
        bitstream = BitstreamDeserialised("id", f"test.{extension.upper()}")
        result = _transform_file_value(bitstream)
        assert isinstance(result, FileValueRDF)
        assert result.prop_type_info.knora_type == KNORA_API.AudioFileValue
        assert result.value == Literal(bitstream.value, datatype=XSD.string)

    @pytest.mark.parametrize("extension", ["pdf", "doc", "docx", "xls", "xlsx", "ppt", "pptx"])
    def test_document_file(self, extension: str) -> None:
        bitstream = BitstreamDeserialised("id", f"test.{extension}")
        result = _transform_file_value(bitstream)
        assert isinstance(result, FileValueRDF)
        assert result.prop_type_info.knora_type == KNORA_API.DocumentFileValue
        assert result.value == Literal(bitstream.value, datatype=XSD.string)
        bitstream = BitstreamDeserialised("id", f"test.{extension.upper()}")
        result = _transform_file_value(bitstream)
        assert isinstance(result, FileValueRDF)
        assert result.prop_type_info.knora_type == KNORA_API.DocumentFileValue
        assert result.value == Literal(bitstream.value, datatype=XSD.string)

    def test_moving_image(self) -> None:
        bitstream = BitstreamDeserialised("id", "test.mp4")
        result = _transform_file_value(bitstream)
        assert isinstance(result, FileValueRDF)
        assert result.prop_type_info.knora_type == KNORA_API.MovingImageFileValue
        assert result.value == Literal(bitstream.value, datatype=XSD.string)

    @pytest.mark.parametrize("extension", ["jpg", "jpeg", "png", "tif", "tiff", "jp2"])
    def test_still_image_file(self, extension: str) -> None:
        bitstream = BitstreamDeserialised("id", f"test.{extension}")
        result = _transform_file_value(bitstream)
        assert isinstance(result, FileValueRDF)
        assert result.prop_type_info.knora_type == KNORA_API.StillImageFileValue
        assert result.value == Literal(bitstream.value, datatype=XSD.string)
        bitstream = BitstreamDeserialised("id", f"test.{extension.upper()}")
        result = _transform_file_value(bitstream)
        assert isinstance(result, FileValueRDF)
        assert result.prop_type_info.knora_type == KNORA_API.StillImageFileValue
        assert result.value == Literal(bitstream.value, datatype=XSD.string)

    def test_still_image_iiif(self) -> None:
        iiif = IIIFUriDeserialised(
            "id", "https://iiif.wellcomecollection.org/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/full/max/0/default.jpg"
        )
        result = _transform_file_value(iiif)
        assert isinstance(result, FileValueRDF)
        assert result.prop_type_info.knora_type == KNORA_API.StillImageExternalFileValue
        assert result.value == Literal(iiif.value, datatype=XSD.anyURI)

    @pytest.mark.parametrize("extension", ["odd", "rng", "txt", "xml", "xsd", "xsl", "csv", "json"])
    def test_text_file(self, extension: str) -> None:
        bitstream = BitstreamDeserialised("id", f"test.{extension}")
        result = _transform_file_value(bitstream)
        assert isinstance(result, FileValueRDF)
        assert result.prop_type_info.knora_type == KNORA_API.TextFileValue
        assert result.value == Literal(bitstream.value, datatype=XSD.string)
        bitstream = BitstreamDeserialised("id", f"test.{extension.upper()}")
        result = _transform_file_value(bitstream)
        assert isinstance(result, FileValueRDF)
        assert result.prop_type_info.knora_type == KNORA_API.TextFileValue
        assert result.value == Literal(bitstream.value, datatype=XSD.string)

    def test_other(self) -> None:
        bitstream = BitstreamDeserialised("id", "test.other")
        result = _transform_file_value(bitstream)
        assert not result


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("test.jpg", "jpg"),
        ("test.JPG", "jpg"),
        (None, ""),
        ("test", ""),
    ],
)
def test_get_file_extension(value: str | None, expected: str) -> None:
    assert _get_file_extension(value) == expected


if __name__ == "__main__":
    pytest.main([__file__])



import importlib.resources
from copy import deepcopy
from pathlib import Path

from lxml import etree
from rdflib import RDF
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.commands.validate_data.api_clients import ListClient
from dsp_tools.commands.validate_data.api_clients import OntologyClient
from dsp_tools.commands.validate_data.api_clients import ShaclValidator
from dsp_tools.commands.validate_data.api_connection import ApiConnection
from dsp_tools.commands.validate_data.deserialise_input import deserialise_xml
from dsp_tools.commands.validate_data.make_data_rdf import make_data_rdf
from dsp_tools.commands.validate_data.models.data_deserialised import ProjectDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import XMLProject
from dsp_tools.commands.validate_data.models.data_rdf import DataRDF
from dsp_tools.commands.validate_data.models.input_problems import UnknownClassesInData
from dsp_tools.commands.validate_data.models.validation import RDFGraphs
from dsp_tools.commands.validate_data.models.validation import ValidationReportGraphs
from dsp_tools.commands.validate_data.reformat_validaton_result import reformat_validation_graph
from dsp_tools.commands.validate_data.sparql.construct_shacl import construct_shapes_graphs
from dsp_tools.commands.validate_data.utils import reformat_onto_iri
from dsp_tools.models.exceptions import InputError
from dsp_tools.utils.ansi_colors import BACKGROUND_BOLD_GREEN
from dsp_tools.utils.ansi_colors import BACKGROUND_BOLD_MAGENTA
from dsp_tools.utils.ansi_colors import BACKGROUND_BOLD_YELLOW
from dsp_tools.utils.ansi_colors import BOLD_CYAN
from dsp_tools.utils.ansi_colors import RESET_TO_DEFAULT
from dsp_tools.utils.xml_utils import parse_xml_file
from dsp_tools.utils.xml_utils import remove_comments_from_element_tree
from dsp_tools.utils.xml_utils import transform_special_tags_make_localname
from dsp_tools.utils.xml_validation import validate_xml

LIST_SEPARATOR = "\n    - "
KNORA_API = "http://api.knora.org/ontology/knora-api/v2#"


VALIDATION_ERRORS_FOUND_MSG = BACKGROUND_BOLD_MAGENTA + "\n   Validation errors found!   " + RESET_TO_DEFAULT


def validate_data(filepath: Path, api_url: str, dev_route: bool, save_graphs: bool) -> bool:  # noqa: ARG001 (unused argument)
    """
    Takes a file and project information and validates it against the ontologies on the server.

    Args:
        filepath: path to the xml data file
        api_url: url of the api host
        dev_route: if this flag is set features that are still in development will be used
        save_graphs: if this flag is set, all the graphs will be saved in a folder

    Returns:
        true unless it crashed
    """
    _inform_about_experimental_feature()
    api_con = ApiConnection(api_url)
    graphs = _get_parsed_graphs(api_con, filepath)
    if unknown_classes := _check_for_unknown_resource_classes(graphs):
        msg = unknown_classes.get_msg()
        print(VALIDATION_ERRORS_FOUND_MSG)
        print(msg)
        return True
    report = _get_validation_result(graphs, api_con, filepath, save_graphs)
    if report.conforms:
        print(BACKGROUND_BOLD_GREEN + "\n   Validation passed!   " + RESET_TO_DEFAULT)
    else:
        reformatted = reformat_validation_graph(report)
        problem_msg = reformatted.get_msg(filepath)
        print(VALIDATION_ERRORS_FOUND_MSG)
        print(problem_msg)
        if reformatted.unexpected_results:
            if save_graphs:
                print(
                    BACKGROUND_BOLD_YELLOW + "\n   Unexpected violations were found! "
                    "Consult the saved graphs for details.   " + RESET_TO_DEFAULT
                )
                return True
            reformatted.unexpected_results.save_inform_user(
                results_graph=report.validation_graph,
                shacl=report.shacl_graph,
                data=report.data_graph,
            )
    return True


def _inform_about_experimental_feature() -> None:
    what_is_validated = [
        "This is an experimental feature, it will change and be extended continuously. "
        "Please note that special characters may not be rendered correctly in the validation result. "
        "This however has no influence on the validation itself."
        "The following information of your data is being validated:",
        "Cardinalities",
        "If the value type used matches the ontology",
        "Content of the values",
        "Missing files",
        "If the file type matches the ontology",
        "DSP in-built resources: link (LinkObj)",
    ]
    print(BOLD_CYAN + LIST_SEPARATOR.join(what_is_validated) + RESET_TO_DEFAULT)


def _get_parsed_graphs(api_con: ApiConnection, filepath: Path) -> RDFGraphs:
    data_rdf, shortcode = _get_data_info_from_file(filepath, api_con.api_url)
    onto_client = OntologyClient(api_con, shortcode)
    list_client = ListClient(api_con, shortcode)
    rdf_graphs = _create_graphs(onto_client, list_client, data_rdf)
    return rdf_graphs


def _check_for_unknown_resource_classes(rdf_graphs: RDFGraphs) -> UnknownClassesInData | None:
    used_cls = _get_all_used_classes(rdf_graphs.data)
    res_cls, value_cls = _get_all_onto_classes(rdf_graphs.ontos + rdf_graphs.knora_api)
    all_cls = res_cls.union(value_cls)
    if extra_cls := used_cls - all_cls:
        return UnknownClassesInData(unknown_classes=extra_cls, classes_onto=res_cls)
    return None


def _get_all_used_classes(data_graph: Graph) -> set[str]:
    types_used = set(data_graph.objects(predicate=RDF.type))
    return {reformat_onto_iri(x) for x in types_used}


def _get_all_onto_classes(ontos: Graph) -> tuple[set[str], set[str]]:
    is_resource_iri = URIRef(KNORA_API + "isResourceClass")
    resource_classes = set(ontos.subjects(is_resource_iri, Literal(True)))
    res_cls = {reformat_onto_iri(x) for x in resource_classes}
    is_value_iri = URIRef(KNORA_API + "isValueClass")
    value_classes = set(ontos.subjects(is_value_iri, Literal(True)))
    value_cls = {reformat_onto_iri(x) for x in value_classes}
    return res_cls, value_cls


def _get_validation_result(
    rdf_graphs: RDFGraphs, api_con: ApiConnection, filepath: Path, save_graphs: bool
) -> ValidationReportGraphs:
    generic_filepath = Path()
    if save_graphs:
        generic_filepath = _save_graphs(filepath, rdf_graphs)
    val = ShaclValidator(api_con, rdf_graphs)
    report = _validate(val)
    if save_graphs:
        report.validation_graph.serialize(f"{generic_filepath}_VALIDATION_REPORT.ttl")
    return report


def _create_graphs(onto_client: OntologyClient, list_client: ListClient, data_rdf: DataRDF) -> RDFGraphs:
    ontologies = _get_project_ontos(onto_client)
    all_lists = list_client.get_lists()
    knora_ttl = onto_client.get_knora_api()
    knora_api = Graph()
    knora_api.parse(data=knora_ttl, format="ttl")
    onto_for_construction = deepcopy(ontologies) + knora_api
    shapes = construct_shapes_graphs(onto_for_construction, all_lists)
    api_shapes = Graph()
    api_shapes_path = importlib.resources.files("dsp_tools").joinpath("resources/validate_data/api-shapes.ttl")
    api_shapes.parse(str(api_shapes_path))
    file_shapes = Graph()
    file_shapes_path = importlib.resources.files("dsp_tools").joinpath(
        "resources/validate_data/api-shapes-with-cardinalities.ttl"
    )
    file_shapes.parse(str(file_shapes_path))
    content_shapes = shapes.content + api_shapes
    card_shapes = shapes.cardinality + file_shapes
    data = data_rdf.make_graph()
    return RDFGraphs(
        data=data,
        ontos=ontologies,
        cardinality_shapes=card_shapes,
        content_shapes=content_shapes,
        knora_api=knora_api,
    )


def _get_project_ontos(onto_client: OntologyClient) -> Graph:
    all_ontos = onto_client.get_ontologies()
    onto_g = Graph()
    for onto in all_ontos:
        og = Graph()
        og.parse(data=onto, format="ttl")
        onto_g += og
    return onto_g


def _save_graphs(filepath: Path, rdf_graphs: RDFGraphs) -> Path:
    parent_directory = filepath.parent
    new_directory = parent_directory / "graphs"
    new_directory.mkdir(exist_ok=True)
    print(BOLD_CYAN + f"\n   Saving graphs to {new_directory}   " + RESET_TO_DEFAULT)
    generic_filepath = new_directory / filepath.stem
    rdf_graphs.ontos.serialize(f"{generic_filepath}_ONTO.ttl")
    shacl_onto = rdf_graphs.content_shapes + rdf_graphs.cardinality_shapes + rdf_graphs.ontos
    shacl_onto.serialize(f"{generic_filepath}_SHACL_ONTO.ttl")
    rdf_graphs.cardinality_shapes.serialize(f"{generic_filepath}_SHACL_CARD.ttl")
    rdf_graphs.content_shapes.serialize(f"{generic_filepath}_SHACL_CONTENT.ttl")
    rdf_graphs.data.serialize(f"{generic_filepath}_DATA.ttl")
    onto_data = rdf_graphs.data + rdf_graphs.ontos
    onto_data.serialize(f"{generic_filepath}_ONTO_DATA.ttl")
    return generic_filepath


def _validate(validator: ShaclValidator) -> ValidationReportGraphs:
    validation_results = validator.validate()
    return ValidationReportGraphs(
        conforms=validation_results.conforms,
        validation_graph=validation_results.validation_graph,
        shacl_graph=validator.rdf_graphs.cardinality_shapes + validator.rdf_graphs.content_shapes,
        onto_graph=validator.rdf_graphs.ontos + validator.rdf_graphs.knora_api,
        data_graph=validator.rdf_graphs.data,
    )


def _get_data_info_from_file(file: Path, api_url: str) -> tuple[DataRDF, str]:
    xml_project = _parse_and_clean_file(file, api_url)
    deserialised: ProjectDeserialised = deserialise_xml(xml_project.root)
    rdf_data: DataRDF = make_data_rdf(deserialised.data)
    return rdf_data, deserialised.info.shortcode


def _parse_and_clean_file(file: Path, api_url: str) -> XMLProject:
    root = parse_xml_file(file)
    root = remove_comments_from_element_tree(root)
    validate_xml(root)
    root = transform_special_tags_make_localname(root)
    return _replace_namespaces(root, api_url)


def _replace_namespaces(root: etree._Element, api_url: str) -> XMLProject:
    new_root = deepcopy(root)
    shortcode = root.attrib["shortcode"]
    default_ontology = root.attrib["default-ontology"]
    namespace_lookup = _make_namespace_lookup(api_url, shortcode, default_ontology)
    for ele in new_root.iterdescendants():
        if (found := ele.attrib.get("restype")) or (found := ele.attrib.get("name")):
            split_found = found.split(":")
            if len(split_found) == 1:
                ele.set("restype" if "restype" in ele.attrib else "name", f"{KNORA_API}{found}")
            elif len(split_found) == 2:
                if len(split_found[0]) == 0:
                    found_namespace = namespace_lookup[default_ontology]
                elif not (namespace := namespace_lookup.get(split_found[0])):
                    found_namespace = _construct_namespace(api_url, shortcode, split_found[0])
                    namespace_lookup[split_found[0]] = found_namespace
                else:
                    found_namespace = namespace
                ele.set("restype" if "restype" in ele.attrib else "name", f"{found_namespace}{split_found[1]}")
            else:
                raise InputError(
                    f"It is not permissible to have a colon in a property or resource class name. "
                    f"Please correct the following: {found}"
                )
    return XMLProject(
        shortcode=shortcode,
        root=new_root,
        used_ontologies=set(namespace_lookup.values()),
    )


def _make_namespace_lookup(api_url: str, shortcode: str, default_onto: str) -> dict[str, str]:
    return {default_onto: _construct_namespace(api_url, shortcode, default_onto), "knora-api": KNORA_API}


def _construct_namespace(api_url: str, shortcode: str, onto_name: str) -> str:
    return f"{api_url}/ontology/{shortcode}/{onto_name}/v2#"
